#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from abc import abstractmethod
from collections.abc import Callable
import math
import operator
import datetime
from calendar import isleap
from decimal import Decimal, Context
from typing import cast, Any, TypeVar, Union

from elementpath.aliases import XPathParserType
from elementpath.helpers import MONTH_DAYS_LEAP, MONTH_DAYS, DAYS_IN_4Y, \
    DAYS_IN_100Y, DAYS_IN_400Y, days_from_common_era, adjust_day, \
    normalized_seconds, months2days, round_number, LazyPattern
from .any_types import AnyAtomicType
from .untyped import UntypedAtomic

###
# Constants for base delta operations
_ZERO_DELTA = datetime.timedelta(0)
_1DAY_DELTA = datetime.timedelta(days=1)
_REF_DATETIME = datetime.datetime(1, 1, 1)
_MAX_OFFSET = datetime.timedelta(hours=14, minutes=0)
_MIN_OFFSET = datetime.timedelta(hours=-14, minutes=0)

__all__ = ['Timezone', 'AbstractDateTime', 'DateTime', 'GregorianDay',
           'GregorianMonth', 'GregorianYear', 'GregorianMonthDay',
           'Date', 'GregorianDay', 'GregorianMonthDay', 'Duration',
           'DayTimeDuration', 'YearMonthDuration', 'GregorianYearMonth',
           'GregorianYearMonth10', 'DateTime10', 'GregorianYear10', 'Time',
           'DateTimeStamp', 'Date10']


class Timezone(datetime.tzinfo):
    """
    A tzinfo implementation for XSD timezone offsets. Offsets must be specified
    between -14:00 and +14:00.

    :param offset: a timedelta instance or an XSD timezone formatted string.
    """
    def __init__(self, offset: datetime.timedelta) -> None:
        super().__init__()
        if not isinstance(offset, datetime.timedelta):
            raise TypeError("offset must be a datetime.timedelta")
        if offset < _MIN_OFFSET or offset > _MAX_OFFSET:
            raise ValueError("offset must be between -14:00 and +14:00")
        self.offset = offset

    @classmethod
    def fromstring(cls, text: str) -> 'Timezone':
        try:
            hours, minutes = text.strip().split(':')
            if hours.startswith('-'):
                return cls(datetime.timedelta(hours=int(hours), minutes=-int(minutes)))
            else:
                return cls(datetime.timedelta(hours=int(hours), minutes=int(minutes)))
        except AttributeError:
            raise TypeError("argument is not a string")
        except ValueError:
            if text.strip() == 'Z':
                return cls(datetime.timedelta(0))
            raise ValueError("%r: not an XSD timezone formatted string" % text) from None

    @classmethod
    def fromduration(cls, duration: 'Duration') -> 'Timezone':
        if duration.seconds % 60 != 0:
            raise ValueError("{!r} has not an integral number of minutes".format(duration))
        return cls(datetime.timedelta(seconds=int(duration.seconds)))

    def __getinitargs__(self) -> tuple[datetime.timedelta]:
        return self.offset,

    def __hash__(self) -> int:
        return hash(self.offset)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Timezone) and self.offset == other.offset

    def __ne__(self, other: object) -> bool:
        return not isinstance(other, Timezone) or self.offset != other.offset

    def __repr__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self.offset)

    def __str__(self) -> str:
        return self.tzname(None)

    def utcoffset(self, dt: datetime.datetime | None) -> datetime.timedelta:
        if not isinstance(dt, datetime.datetime) and dt is not None:
            raise TypeError("utcoffset() argument must be a "
                            "datetime.datetime instance or None")
        return self.offset

    def tzname(self, dt: datetime.datetime | None) -> str:
        if not isinstance(dt, datetime.datetime) and dt is not None:
            raise TypeError("tzname() argument must be a "
                            "datetime.datetime instance or None")

        if not self.offset:
            return 'Z'
        elif self.offset < datetime.timedelta(0):
            sign, offset = '-', -self.offset
        else:
            sign, offset = '+', self.offset

        hours, minutes = offset.seconds // 3600, offset.seconds // 60 % 60
        return '{}{:02d}:{:02d}'.format(sign, hours, minutes)

    def dst(self, dt: datetime.datetime | None) -> None:
        if not isinstance(dt, datetime.datetime) and dt is not None:
            raise TypeError("dst() argument must be a "
                            "datetime.datetime instance or None")

    def fromutc(self, dt: datetime.datetime) -> datetime.datetime:
        if isinstance(dt, datetime.datetime):
            return dt + self.offset
        raise TypeError("fromutc() argument must be a datetime.datetime instance")


_UTC_TIMEZONE = Timezone(datetime.timedelta(0))


def get_comparable_datetimes(dt1: datetime.datetime, dt2: datetime.datetime) \
        -> tuple[datetime.datetime, datetime.datetime]:
    """
    Fill timezone because Python can't compare offset-naive and offset-aware datetimes.
    """
    if dt1.tzinfo is dt2.tzinfo:
        return dt1, dt2
    elif dt1.tzinfo is None:
        return dt1.replace(tzinfo=_UTC_TIMEZONE), dt2
    elif dt2.tzinfo is None:
        return dt1, dt2.replace(tzinfo=_UTC_TIMEZONE)
    else:
        return dt1, dt2


DT = TypeVar('DT', bound='AbstractDateTime')


class AbstractDateTime(AnyAtomicType):
    """
    A class for representing XSD date/time objects. It uses and internal datetime.datetime
    attribute and an integer attribute for processing BCE years or for years after 9999 CE.
    """
    pattern = LazyPattern(r'^$')

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'AbstractDateTime':
        match value:
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case datetime.datetime():
                return cls.fromdatetime(value)
            case _:
                return cls.fromstring(value)

    __slots__ = ('_dt', '_year')

    def __init__(self, year: int = 2000, month: int = 1, day: int = 1, hour: int = 0,
                 minute: int = 0, second: int = 0, microsecond: int = 0,
                 tzinfo: datetime.tzinfo | None = None) -> None:

        if hour == 24 and minute == second == microsecond == 0:
            hour = 0
            if year == 9999 and month == 12 and day == 31:
                delta = _ZERO_DELTA
                year = 10000
                month = 1
                day = 1
            else:
                delta = _1DAY_DELTA
                hour = 0
        else:
            delta = _ZERO_DELTA

        if 1 <= year <= 9999:
            self._dt = datetime.datetime(year, month, day, hour, minute,
                                         second, microsecond, tzinfo)
            if delta:
                self._dt += delta
            self._year = self._dt.year
        elif year == 0:
            raise ValueError('0 is an illegal value for year')
        elif not isinstance(year, int):
            raise TypeError("invalid type %r for year" % type(year))
        elif abs(year) > 2 ** 31:
            raise OverflowError("year overflow")
        else:
            self._year = year
            if isleap(year + bool(self._xsd_version != '1.0')):
                self._dt = datetime.datetime(4, month, day, hour, minute,
                                             second, microsecond, tzinfo)
            else:
                self._dt = datetime.datetime(6, month, day, hour, minute,
                                             second, microsecond, tzinfo)
            if delta:
                self._dt += delta

    def __repr__(self) -> str:
        fields = self.pattern.groupindex.keys()
        arg_string = ', '.join(
            str(getattr(self, k))
            for k in ['year', 'month', 'day', 'hour', 'minute'] if k in fields
        )
        if 'second' in fields:
            if self.microsecond:
                arg_string += ', %d.%06d' % (self.second, self.microsecond)
            else:
                arg_string += ', %d' % self.second

        if self.tzinfo is not None:
            arg_string += ', tzinfo=%r' % self.tzinfo
        return '%s(%s)' % (self.__class__.__name__, arg_string)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash((self._dt, self._year))

    def __eq__(self, other: object) -> bool:
        return self._compare(other, operator.eq)

    def __ne__(self, other: object) -> bool:
        return not self._compare(other, operator.eq)

    def __lt__(self, other: object) -> bool:
        return self._compare(other, operator.lt)

    def __le__(self, other: object) -> bool:
        return self._compare(other, operator.le)

    def __gt__(self, other: object) -> bool:
        return self._compare(other, operator.gt)

    def __ge__(self, other: object) -> bool:
        return self._compare(other, operator.ge)

    def __add__(self, other: object) -> Union['DayTimeDuration', 'AbstractDateTime']:
        if isinstance(other, AbstractDateTime):
            raise TypeError("wrong type %r for operand %r" % (type(other), other))
        return self._operation(other, operator.add)

    def __sub__(self, other: object) -> Union['DayTimeDuration', 'AbstractDateTime']:
        return self._operation(other, operator.sub)

    def _compare(self, other: object, op: Callable[[Any, Any], bool]) -> bool:
        if isinstance(other, datetime.datetime):
            dt, year = other, other.year
        elif isinstance(other, AbstractDateTime):
            match (self.name, other.name):
                case ('time', 'date') | ('date', 'time'):
                    if op is operator.eq:
                        raise TypeError("wrong type %r for operand %r" % (type(other), other))
                case ('dateTime', 'date') | ('date', 'dateTime'):
                    if op is not operator.eq:
                        raise TypeError("wrong type %r for operand %r" % (type(other), other))

            dt, year = other._dt, other._year
        elif op is operator.eq:
            return False
        else:
            raise TypeError("wrong type %r for operand %r" % (type(other), other))

        if self._year != year:
            return op(self._year, year)
        elif self._dt.tzinfo is dt.tzinfo:
            return op(self._dt, dt)
        elif self.tzinfo is None:
            return op(self._dt.replace(tzinfo=_UTC_TIMEZONE), dt)
        elif other.tzinfo is None:
            return op(self._dt, dt.replace(tzinfo=_UTC_TIMEZONE))
        else:
            return op(self._dt, dt)

    def _operation(self, other: object, op: Callable[[Any, Any], Any]) \
            -> Union['DayTimeDuration', 'AbstractDateTime']:
        match other:
            case AbstractDateTime():
                dt1, dt2 = get_comparable_datetimes(self._dt, other._dt)
                if 1 <= self._year <= 9999 and 1 <= other._year <= 9999:
                    return DayTimeDuration.fromtimedelta(dt1 - dt2)
                return DayTimeDuration.fromtimedelta(self.todelta() - other.todelta())

            case datetime.timedelta():
                delta = op(self.todelta(), other)
                return type(self).fromdelta(delta, adjust_timezone=True)

            case DayTimeDuration():
                delta = op(self.todelta(), other.get_timedelta())
                tzinfo = self._dt.tzinfo
                if tzinfo is None:
                    return type(self).fromdelta(delta)

                value = type(self).fromdelta(delta + tzinfo.utcoffset(None))
                value.tzinfo = tzinfo
                return value

            case YearMonthDuration():
                month = op(self._dt.month - 1, other.months) % 12 + 1
                year = self._year + op(self._dt.month - 1, other.months) // 12
                day = adjust_day(year, month, self._dt.day)

                if year > 0:
                    dt = self._dt.replace(year=year, month=month, day=day)
                elif isleap(year):
                    dt = self._dt.replace(year=4, month=month, day=day)
                else:
                    dt = self._dt.replace(year=6, month=month, day=day)

                kwargs = {k: getattr(dt, k) for k in self.pattern.groupindex.keys()}
                if year <= 0:
                    kwargs['year'] = year
                return type(self)(**kwargs)

            case _:
                raise TypeError("wrong type %r for operand %r" % (type(other), other))

    @property
    def year(self) -> int:
        return self._year

    @property
    def bce(self) -> bool:
        return self._year < 0

    @property
    def iso_year(self) -> str:
        """The ISO string representation of the year field."""
        year = self.year
        if -9999 <= year < -1:
            return '{:05}'.format(year if self._xsd_version == '1.0' else year + 1)
        elif year == -1:
            return '-0001' if self._xsd_version == '1.0' else '0000'
        elif 0 <= year <= 9999:
            return '{:04}'.format(year)
        else:
            return str(year)

    @property
    def month(self) -> int:
        return self._dt.month

    @property
    def day(self) -> int:
        return self._dt.day

    @property
    def hour(self) -> int:
        return self._dt.hour

    @property
    def minute(self) -> int:
        return self._dt.minute

    @property
    def second(self) -> int:
        return self._dt.second

    @property
    def microsecond(self) -> int:
        return self._dt.microsecond

    @property
    def tzinfo(self) -> datetime.tzinfo | None:
        return cast(Timezone, self._dt.tzinfo)

    @tzinfo.setter
    def tzinfo(self, tz: datetime.tzinfo | None) -> None:
        self._dt = self._dt.replace(tzinfo=tz)

    def tzname(self) -> str | None:
        return self._dt.tzname()

    def astimezone(self, tz: datetime.tzinfo | None = None) -> datetime.datetime:
        return self._dt.astimezone(tz)

    def isocalendar(self) -> tuple[int, int, int]:
        return cast(tuple[int, int, int], self._dt.isocalendar())

    @classmethod
    def fromstring(cls: type[DT], datetime_string: str,
                   tzinfo: datetime.tzinfo | None = None) -> DT:
        """
        Creates an XSD date/time instance from a string formatted value.

        :param datetime_string: a string containing an XSD formatted date/time specification.
        :param tzinfo: optional implicit timezone information (defaults to UTC).
        :return: an AbstractDateTime concrete subclass instance.
        """
        if not isinstance(datetime_string, str):
            msg = '1st argument has an invalid type {!r}'
            raise TypeError(msg.format(type(datetime_string)))
        elif tzinfo and not isinstance(tzinfo, Timezone):
            msg = '2nd argument has an invalid type {!r}'
            raise TypeError(msg.format(type(tzinfo)))

        match = cls.pattern.match(datetime_string.strip())
        if match is None:
            msg = 'Invalid datetime string {!r} for {!r}'
            raise ValueError(msg.format(datetime_string, cls))

        match_dict = match.groupdict()
        kwargs: dict[str, int] = {
            k: int(v) for k, v in match_dict.items() if k != 'tzinfo' and v is not None
        }

        if match_dict['tzinfo'] is not None:
            tzinfo = Timezone.fromstring(match_dict['tzinfo'])

        if 'microsecond' in kwargs:
            microseconds = match_dict['microsecond']
            if len(microseconds) != 6:
                microseconds += '0' * (6 - len(microseconds))
                kwargs['microsecond'] = int(microseconds[:6])

        if 'year' in kwargs:
            year_digits = match_dict['year'].lstrip('-')
            if year_digits.startswith('0') and len(year_digits) > 4:
                msg = "Invalid datetime string {!r} for {!r} (when year " \
                      "exceeds 4 digits leading zeroes are not allowed)"
                raise ValueError(msg.format(datetime_string, cls))

            if cls._xsd_version == '1.0':
                if kwargs['year'] == 0:
                    raise ValueError("year '0000' is an illegal value for XSD 1.0")
            elif kwargs['year'] <= 0:
                kwargs['year'] -= 1

        return cls(tzinfo=tzinfo, **kwargs)

    @classmethod
    def fromdatetime(cls: type[DT], dt: Union[datetime.datetime, datetime.date, datetime.time],
                     year: int | None = None) -> DT:
        """
        Creates an XSD date/time instance from a datetime.datetime/date/time instance.

        :param dt: the datetime, date or time instance that stores the XSD Date/Time value.
        :param year: if a year is provided the created instance refers to it and the \
        possibly present *dt.year* part is ignored.
        :return: an AbstractDateTime concrete subclass instance.
        """
        if not isinstance(dt, (datetime.datetime, datetime.date, datetime.time)):
            raise TypeError('1st argument has an invalid type %r' % type(dt))
        elif year is not None and not isinstance(year, int):
            raise TypeError('2nd argument has an invalid type %r' % type(year))

        kwargs = {k: getattr(dt, k) for k in cls.pattern.groupindex.keys() if hasattr(dt, k)}
        if year is not None:
            kwargs['year'] = year
        return cls(**kwargs)

    @classmethod
    def fromdelta(cls, delta: datetime.timedelta, adjust_timezone: bool = False) \
            -> 'AbstractDateTime':
        """
        Creates an XSD dateTime/date instance from a datetime.timedelta related to
        0001-01-01T00:00:00 CE. In case of a date the time part is not counted.

        :param delta: a datetime.timedelta instance.
        :param adjust_timezone: if `True` adjusts the timezone of Date objects \
        with eventually present hours and minutes.
        """
        try:
            dt = _REF_DATETIME + delta
        except OverflowError:
            days = delta.days
            if days > 0:
                y400, days = divmod(days, DAYS_IN_400Y)
                y100, days = divmod(days, DAYS_IN_100Y)
                y4, days = divmod(days, DAYS_IN_4Y)
                y1, days = divmod(days, 365)
                year = y400 * 400 + y100 * 100 + y4 * 4 + y1 + 1
                if y1 == 4 or y100 == 4:
                    year -= 1
                    days = 365

                td = datetime.timedelta(days=days, seconds=delta.seconds,
                                        microseconds=delta.microseconds)
                dt = datetime.datetime(4 if isleap(year) else 6, 1, 1) + td

            elif days >= -366:
                year = -1
                td = datetime.timedelta(days=days, seconds=delta.seconds,
                                        microseconds=delta.microseconds)
                dt = datetime.datetime(5, 1, 1) + td

            else:
                days = -days - 366
                y400, days = divmod(days, DAYS_IN_400Y)
                y100, days = divmod(days, DAYS_IN_100Y)
                y4, days = divmod(days, DAYS_IN_4Y)
                y1, days = divmod(days, 365)
                year = -y400 * 400 - y100 * 100 - y4 * 4 - y1 - 2
                if y1 == 4 or y100 == 4:
                    year += 1
                    days = 365

                td = datetime.timedelta(days=-days, seconds=delta.seconds,
                                        microseconds=delta.microseconds)
                if not td:
                    dt = datetime.datetime(4 if isleap(year + 1) else 6, 1, 1)
                    year += 1
                else:
                    dt = datetime.datetime(5 if isleap(year + 1) else 7, 1, 1) + td
        else:
            year = dt.year

        if issubclass(cls, Date):
            if adjust_timezone and (dt.hour or dt.minute):
                assert dt.tzinfo is None
                hour, minute = dt.hour, dt.minute

                if hour < 14 or hour == 14 and minute == 0:
                    tz = Timezone(datetime.timedelta(hours=-hour, minutes=-minute))
                    dt = dt.replace(tzinfo=tz)
                else:
                    tz = Timezone(datetime.timedelta(hours=-dt.hour + 24, minutes=-minute))
                    dt = dt.replace(tzinfo=tz)
                    dt += datetime.timedelta(days=1)

            return cls(year, dt.month, dt.day, tzinfo=dt.tzinfo)
        return cls(year, dt.month, dt.day, dt.hour, dt.minute,
                   dt.second, dt.microsecond, dt.tzinfo)

    def todelta(self) -> datetime.timedelta:
        """Returns the datetime.timedelta from 0001-01-01T00:00:00 CE."""
        if 1 <= self._year <= 9999:
            if self._dt.tzinfo is None:
                return self._dt - _REF_DATETIME
            else:
                return self._dt - _REF_DATETIME.replace(tzinfo=_UTC_TIMEZONE)

        year, dt = self.year, self._dt
        tzinfo = None if dt.tzinfo is None else _UTC_TIMEZONE

        if year > 0:
            m_days = MONTH_DAYS_LEAP if isleap(year) else MONTH_DAYS
            days = days_from_common_era(year - 1) + sum(m_days[m] for m in range(1, dt.month))
        else:
            m_days = MONTH_DAYS_LEAP if isleap(year + 1) else MONTH_DAYS
            days = days_from_common_era(year) + sum(m_days[m] for m in range(1, dt.month))

        delta = (dt - datetime.datetime(dt.year, dt.month, day=1, tzinfo=tzinfo))
        return datetime.timedelta(days=days, seconds=delta.total_seconds())


class DateTime(AbstractDateTime):
    name = 'dateTime'
    pattern = LazyPattern(
        r'^(?P<year>-?[0-9]*[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
        r'(T(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):'
        r'(?P<second>[0-9]{2})(?:\.(?P<microsecond>[0-9]+))?)'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )
    _xsd_version = '1.1'

    @classmethod
    def make(cls, value: Any,
             parser: XPathParserType | None = None,
             **kwargs: Any) -> 'DateTime':

        if cls._xsd_version == '1.0':
            dt_class = cls
        elif parser is not None:
            dt_class = DateTime10 if parser.xsd_version == '1.0' else cls
        elif kwargs.get('xsd_version') == '1.0':
            dt_class = DateTime10
        else:
            dt_class = cls

        match value:
            case UntypedAtomic():
                return dt_class.fromstring(value.value)
            case Date():
                return dt_class(value.year, value.month, value.day, tzinfo=value.tzinfo)
            case DateTime():
                if value.__class__ is dt_class:
                    return value

                year = value.year
                if year < 0:
                    if value.xsd_versions != '1.0':
                        year += 1
                    if not dt_class.__name__.endswith('10'):
                        year -= 1

                return dt_class(
                    year, value.month, value.day, value.hour, value.minute,
                    value.second, value.microsecond, value.tzinfo
                )
            case _:
                return dt_class.fromstring(value)

    def __init__(self, year: int, month: int, day: int,
                 hour: int = 0, minute: int = 0, second: int = 0,
                 microsecond: int = 0,
                 tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param year: the year, between -9999 and 9999
        :param month: the month, between 1 and 12
        :param day: the day, between 1 and 31
        :param hour: the hour, between 0 and 23
        :param minute: the minute, between 0 and 59
        :param second: the second, between 0 and 59
        :param microsecond: the microsecond, between 0 and 999999
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(year, month, day, hour, minute, second, microsecond, tzinfo)

    def __str__(self) -> str:
        if self.microsecond:
            return '{}-{:02}-{:02}T{:02}:{:02}:{:02}.{}{}'.format(
                self.iso_year, self.month, self.day, self.hour, self.minute, self.second,
                '{:06}'.format(self.microsecond).rstrip('0'), str(self.tzinfo or '')
            )
        return '{}-{:02}-{:02}T{:02}:{:02}:{:02}{}'.format(
            self.iso_year, self.month, self.day, self.hour,
            self.minute, self.second, str(self.tzinfo or '')
        )


class DateTime10(DateTime):
    name = 'dateTime'
    _xsd_version = '1.0'

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return subclass is DateTime10 or subclass is DateTime


class DateTimeStamp(DateTime):
    name = 'dateTimeStamp'
    pattern = LazyPattern(
        r'^(?P<year>-?[0-9]*[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
        r'(T(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):'
        r'(?P<second>[0-9]{2})(?:\.(?P<microsecond>[0-9]+))?)'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))$'
    )


class Date(AbstractDateTime):
    name = 'date'
    pattern = LazyPattern(
        r'^(?P<year>-?[0-9]*[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )
    _xsd_version = '1.1'

    @classmethod
    def make(cls, value: Any,
             parser: XPathParserType | None = None,
             **kwargs: Any) -> 'Date':

        if cls._xsd_version == '1.0':
            dt_class = cls
        elif parser is not None:
            dt_class = Date10 if parser.xsd_version == '1.0' else Date
        elif kwargs.get('xsd_version') == '1.0':
            dt_class = Date10
        else:
            dt_class = cls

        match value:
            case UntypedAtomic():
                return dt_class.fromstring(value.value)
            case DateTime() | Date():
                if isinstance(value, Date) and value.__class__ is dt_class:
                    return value

                year = value.year
                if year < 0:
                    if value.xsd_versions != '1.0':
                        year += 1
                    if dt_class._xsd_version != '1.0':
                        year -= 1

                return dt_class(year, value.month, value.day, value.tzinfo)
            case _:
                return dt_class.fromstring(value)

    def __init__(self, year: int, month: int, day: int,
                 tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param year: the year, between -9999 and 9999
        :param month: the month, between 1 and 12
        :param day: the day, between 1 and 31
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(year, month, day, tzinfo=tzinfo)

    def __str__(self) -> str:
        return '{}-{:02}-{:02}{}'.format(
            self.iso_year, self.month, self.day, str(self.tzinfo or '')
        )


class Date10(Date):
    name = 'date'
    _xsd_version = '1.0'

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return subclass is Date10 or subclass is Date


class GregorianDay(AbstractDateTime):
    name = 'gDay'
    pattern = LazyPattern(
        r'^---(?P<day>[0-9]{2})'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'GregorianDay':
        match value:
            case GregorianDay():
                return value
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case Date() | DateTime():
                return cls(value.day, value.tzinfo)
            case _:
                return cls.fromstring(value)

    def __init__(self, day: int, tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param day: the day, between 1 and 31
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(day=day, tzinfo=tzinfo)

    def __str__(self) -> str:
        return '---{:02}{}'.format(self.day, str(self.tzinfo or ''))


class GregorianMonth(AbstractDateTime):
    name = 'gMonth'
    pattern = LazyPattern(
        r'^--(?P<month>[0-9]{2})'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'GregorianMonth':
        match value:
            case GregorianMonth():
                return value
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case Date() | DateTime():
                return cls(value.month, value.tzinfo)
            case _:
                return cls.fromstring(value)

    def __init__(self, month: int, tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param month: the month, between 1 and 12
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(month=month, tzinfo=tzinfo)

    def __str__(self) -> str:
        return '--{:02}{}'.format(self.month, str(self.tzinfo or ''))


class GregorianMonthDay(AbstractDateTime):
    name = 'gMonthDay'
    pattern = LazyPattern(
        r'^--(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'GregorianMonthDay':
        match value:
            case GregorianMonthDay():
                return value
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case Date() | DateTime():
                return cls(value.month, value.day, value.tzinfo)
            case _:
                return cls.fromstring(value)

    def __init__(self, month: int, day: int, tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param month: the month, between 1 and 12
        :param day: the day, between 1 and 31
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(month=month, day=day, tzinfo=tzinfo)

    def __str__(self) -> str:
        return '--{:02}-{:02}{}'.format(self.month, self.day, str(self.tzinfo or ''))


class GregorianYear(AbstractDateTime):
    name = 'gYear'
    pattern = LazyPattern(
        r'^(?P<year>-?[0-9]*[0-9]{4})'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )
    _xsd_version = '1.1'

    @classmethod
    def make(cls, value: Any,
             parser: XPathParserType | None = None,
             **kwargs: Any) -> 'GregorianYear':

        if cls._xsd_version == '1.0':
            dt_class = cls
        elif parser is not None:
            dt_class = GregorianYear10 if parser.xsd_version == '1.0' else GregorianYear
        elif kwargs.get('xsd_version') == '1.0':
            dt_class = GregorianYear10
        else:
            dt_class = GregorianYear

        match value:
            case UntypedAtomic():
                return dt_class.fromstring(value.value)
            case Date() | DateTime() | GregorianYear():
                if value.__class__ is dt_class and isinstance(value, GregorianYear):
                    return value

                year = value.year
                if year < 0:
                    if value.xsd_versions != '1.0':
                        year += 1
                    if dt_class._xsd_version != '1.0':
                        year -= 1

                return dt_class(year, value.tzinfo)
            case _:
                return dt_class.fromstring(value)

    def __init__(self, year: int, tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param year: the year, between -9999 and 9999
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(year, tzinfo=tzinfo)

    def __str__(self) -> str:
        return '{}{}'.format(self.iso_year, str(self.tzinfo or ''))


class GregorianYear10(GregorianYear):
    name = 'gYear'
    _xsd_version = '1.0'

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return subclass is GregorianYear10 or subclass is GregorianYear


class GregorianYearMonth(AbstractDateTime):
    name = 'gYearMonth'
    pattern = LazyPattern(
        r'^(?P<year>-?[0-9]*[0-9]{4})-(?P<month>[0-9]{2})'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )
    _xsd_version = '1.1'

    @classmethod
    def make(cls, value: Any,
             parser: XPathParserType | None = None,
             **kwargs: Any) -> 'GregorianYearMonth':

        if cls._xsd_version == '1.0':
            dt_class = cls
        elif parser is not None:
            dt_class = GregorianYearMonth10 if parser.xsd_version == '1.0' else GregorianYearMonth
        elif kwargs.get('xsd_version') == '1.0':
            dt_class = GregorianYearMonth10
        else:
            dt_class = GregorianYearMonth

        match value:
            case UntypedAtomic():
                return dt_class.fromstring(value.value)
            case Date() | DateTime() | GregorianYearMonth():
                if value.__class__ is dt_class and isinstance(value, GregorianYearMonth):
                    return value

                year = value.year
                if year < 0:
                    if value.xsd_versions != '1.0':
                        year += 1
                    if dt_class._xsd_version != '1.0':
                        year -= 1

                return dt_class(year, value.month, value.tzinfo)
            case _:
                return dt_class.fromstring(value)

    def __init__(self, year: int, month: int, tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param year: the year, between -9999 and 9999
        :param month: the month, between 1 and 12
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        super().__init__(year, month, tzinfo=tzinfo)

    def __str__(self) -> str:
        return '{}-{:02}{}'.format(self.iso_year, self.month, str(self.tzinfo or ''))


class GregorianYearMonth10(GregorianYearMonth):
    name = 'gYearMonth'
    _xsd_version = '1.0'

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return subclass is GregorianYearMonth10 or subclass is GregorianYearMonth


class Time(AbstractDateTime):
    name = 'time'
    pattern = LazyPattern(
        r'^(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):'
        r'(?P<second>[0-9]{2})(?:\.(?P<microsecond>[0-9]+))?'
        r'(?P<tzinfo>Z|[+-](?:(?:0[0-9]|1[0-3]):[0-5][0-9]|14:00))?$'
    )

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'Time':
        match value:
            case Time():
                return value
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case DateTime():
                return cls(value.hour, value.minute, value.second,
                           value.microsecond, value.tzinfo)
            case _:
                return cls.fromstring(value)

    def __init__(self, hour: int = 0, minute: int = 0,
                 second: int = 0, microsecond: int = 0,
                 tzinfo: datetime.tzinfo | None = None) -> None:
        """
        :param hour: the hour, between 0 and 23
        :param minute: the minute, between 0 and 59
        :param second: the second, between 0 and 59
        :param microsecond: the microsecond, between 0 and 999999
        :param tzinfo: optional implicit timezone information (defaults to UTC)
        """
        if hour == 24 and minute == second == microsecond == 0:
            hour = 0
        super().__init__(
            hour=hour, minute=minute, second=second, microsecond=microsecond, tzinfo=tzinfo
        )

    def __str__(self) -> str:
        if self.microsecond:
            return '{:02}:{:02}:{:02}.{}{}'.format(
                self.hour, self.minute, self.second,
                '{:06}'.format(self.microsecond).rstrip('0'),
                str(self.tzinfo or '')
            )
        return '{:02}:{:02}:{:02}{}'.format(
            self.hour, self.minute, self.second, str(self.tzinfo or '')
        )

    def __add__(self, other: object) -> 'Time':
        if isinstance(other, DayTimeDuration):
            dt = self._dt + other.get_timedelta()
        elif isinstance(other, datetime.timedelta):
            dt = self._dt + other
        else:
            raise TypeError("wrong type %r for operand %r" % (type(other), other))
        return Time(dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)

    def __sub__(self, other: object) -> Union['DayTimeDuration', 'Time']:
        if isinstance(other, Time):
            dt1, dt2 = get_comparable_datetimes(self._dt, other._dt)
            return DayTimeDuration.fromtimedelta(dt1 - dt2)
        elif isinstance(other, DayTimeDuration):
            dt = self._dt - other.get_timedelta()
            return Time(dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
        elif isinstance(other, datetime.timedelta):
            dt = self._dt - other
            return Time(dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
        else:
            raise TypeError("wrong type %r for operand %r" % (type(other), other))


_D = TypeVar('_D', bound='Duration')


class Duration(AnyAtomicType):
    name = 'duration'
    pattern = LazyPattern(
        r'^(-)?P(?=[0-9]|T)(?:([0-9]+)Y)?(?:([0-9]+)M)?(?:([0-9]+)D)?'
        r'(?:T(?=[0-9])(?:([0-9]+)H)?(?:([0-9]+)M)?(?:([0-9]+(?:\.[0-9]+)?)S)?)?$'
    )

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'Duration':
        match value:
            case Duration():
                return value
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case _:
                return cls.fromstring(value)

    __slots__ = ('months', 'seconds')

    def __init__(self, months: int = 0, seconds: Union[Decimal, int] = 0) -> None:
        """
        :param months: an integer value that represents years and months.
        :param seconds: a decimal or an integer instance that represents \
        days, hours, minutes, seconds and fractions of seconds.
        """
        if seconds < 0 < months or months < 0 < seconds:
            raise ValueError('signs differ: (months=%d, seconds=%d)' % (months, seconds))
        elif abs(months) > 2 ** 31:
            raise OverflowError("months duration overflow")
        elif abs(seconds) > 2 ** 63:  # type: ignore[operator]
            raise OverflowError("seconds duration overflow")

        self.months = months
        self.seconds = Decimal(seconds).quantize(Decimal('1.000000', context=Context(prec=30)))

    def __repr__(self) -> str:
        return '{}(months={!r}, seconds={})'.format(
            self.__class__.__name__, self.months, normalized_seconds(self.seconds)
        )

    def __str__(self) -> str:
        m = abs(self.months)
        years, months = m // 12, m % 12
        s = self.seconds.copy_abs()
        days = int(s // 86400)
        hours = int(s // 3600 % 24)
        minutes = int(s // 60 % 60)
        seconds = s % 60

        value = '-P' if self.sign else 'P'
        if years or months or days:
            if years:
                value += '%dY' % years
            if months:
                value += '%dM' % months
            if days:
                value += '%dD' % days

        if hours or minutes or seconds:
            value += 'T'
            if hours:
                value += '%dH' % hours
            if minutes:
                value += '%dM' % minutes
            if seconds:
                value += '%sS' % normalized_seconds(seconds)

        elif value[-1] == 'P':
            value += 'T0S'
        return value

    @classmethod
    def fromstring(cls: type[_D], text: str) -> _D:
        """
        Creates a Duration instance from a formatted XSD duration string.

        :param text: an ISO 8601 representation without week fragment and an optional decimal part \
        only for seconds fragment.
        """
        if not isinstance(text, str):
            msg = 'argument has an invalid type {!r}'
            raise TypeError(msg.format(type(text)))

        match = cls.pattern.match(text.strip())
        if match is None:
            raise ValueError('%r is not an xs:duration value' % text)

        sign, y, mo, d, h, mi, s = match.groups()
        seconds = Decimal(s or 0)
        minutes = int(mi or 0) + int(seconds // 60)
        seconds = seconds % 60
        hours = int(h or 0) + minutes // 60
        minutes = minutes % 60
        days = int(d or 0) + hours // 24
        hours = hours % 24
        months = int(mo or 0) + 12 * int(y or 0)

        if sign is None:
            seconds = seconds + (days * 24 + hours) * 3600 + minutes * 60
        else:
            months = -months
            seconds = -seconds - (days * 24 + hours) * 3600 - minutes * 60

        if cls is DayTimeDuration:
            if months:
                raise ValueError('months must be 0 for %r' % cls.__name__)
            return cls(seconds=seconds)
        elif cls is YearMonthDuration:
            if seconds:
                raise ValueError('seconds must be 0 for %r' % cls.__name__)
            return cls(months=months)
        return cls(months=months, seconds=seconds)

    @property
    def sign(self) -> str:
        return '-' if self.months < 0 or self.seconds < 0 else ''

    def _compare_durations(self, other: object, op: Callable[[Any, Any], Any]) -> bool:
        """
        Ordering is defined through comparison of four datetime.datetime values.

        Ref: https://www.w3.org/TR/2012/REC-xmlschema11-2-20120405/#duration
        """
        if not isinstance(other, self.__class__):
            raise TypeError("wrong type %r for operand %r" % (type(other), other))

        m1, s1 = self.months, int(self.seconds)
        m2, s2 = other.months, int(other.seconds)
        ms1, ms2 = int((self.seconds - s1) * 1000000), int((other.seconds - s2) * 1000000)
        return all([
            op(datetime.timedelta(months2days(1696, 9, m1), s1, ms1),
               datetime.timedelta(months2days(1696, 9, m2), s2, ms2)),
            op(datetime.timedelta(months2days(1697, 2, m1), s1, ms1),
               datetime.timedelta(months2days(1697, 2, m2), s2, ms2)),
            op(datetime.timedelta(months2days(1903, 3, m1), s1, ms1),
               datetime.timedelta(months2days(1903, 3, m2), s2, ms2)),
            op(datetime.timedelta(months2days(1903, 7, m1), s1, ms1),
               datetime.timedelta(months2days(1903, 7, m2), s2, ms2)),
        ])

    def __hash__(self) -> int:
        return hash((self.months, self.seconds))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.months == other.months and self.seconds == other.seconds
        elif isinstance(other, UntypedAtomic):
            return self.__eq__(self.fromstring(other.value))
        else:
            return other == (self.months, self.seconds)

    def __ne__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.months != other.months or self.seconds != other.seconds
        elif isinstance(other, UntypedAtomic):
            return self.__ne__(self.fromstring(other.value))
        else:
            return other != (self.months, self.seconds)

    def __lt__(self, other: object) -> bool:
        return self._compare_durations(other, operator.lt)

    def __le__(self, other: object) -> bool:
        return self == other or self._compare_durations(other, operator.le)

    def __gt__(self, other: object) -> bool:
        return self._compare_durations(other, operator.gt)

    def __ge__(self, other: object) -> bool:
        return self == other or self._compare_durations(other, operator.ge)


class YearMonthDuration(Duration):
    name = 'yearMonthDuration'

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'YearMonthDuration':
        match value:
            case YearMonthDuration():
                return value
            case Duration():
                return YearMonthDuration(months=value.months)
            case UntypedAtomic():
                return YearMonthDuration.fromstring(value.value)
            case _:
                return YearMonthDuration.fromstring(value)

    def __init__(self, months: int = 0) -> None:
        """
        :param months: an integer value that represents years and months.
        """
        super().__init__(months, 0)

    def __repr__(self) -> str:
        return '%s(months=%r)' % (self.__class__.__name__, self.months)

    def __str__(self) -> str:
        m = abs(self.months)
        years, months = m // 12, m % 12

        if not years:
            return '-P%dM' % months if self.months < 0 else 'P%dM' % months
        elif not months:
            return '-P%dY' % years if self.months < 0 else 'P%dY' % years
        elif self.months < 0:
            return '-P%dY%dM' % (years, months)
        else:
            return 'P%dY%dM' % (years, months)

    def __add__(self, other: object) \
            -> Union['YearMonthDuration', 'DayTimeDuration', 'AbstractDateTime']:
        if isinstance(other, self.__class__):
            return YearMonthDuration(months=self.months + other.months)
        elif isinstance(other, (DateTime, Date)):
            return other + self
        raise TypeError("cannot add %r to %r" % (type(other), type(self)))

    def __sub__(self, other: object) -> 'YearMonthDuration':
        if not isinstance(other, self.__class__):
            raise TypeError("cannot subtract %r from %r" % (type(other), type(self)))
        return YearMonthDuration(months=self.months - other.months)

    def __mul__(self, other: object) -> 'YearMonthDuration':
        if not isinstance(other, (float, int, Decimal)):
            raise TypeError("cannot multiply a %r by %r" % (type(self), type(other)))
        return YearMonthDuration(months=int(round_number(self.months * other)))

    def __truediv__(self, other: object) -> Union[float, 'YearMonthDuration']:
        if isinstance(other, self.__class__):
            return self.months / other.months
        elif isinstance(other, (float, int, Decimal)):
            return YearMonthDuration(months=int(round_number(self.months / other)))
        else:
            raise TypeError("cannot divide a %r by %r" % (type(self), type(other)))


class DayTimeDuration(Duration):
    name = 'dayTimeDuration'

    @classmethod
    def make(cls, value: Any, **kwargs: Any) -> 'DayTimeDuration':
        match value:
            case DayTimeDuration():
                return value
            case Duration():
                return cls(seconds=value.seconds)
            case UntypedAtomic():
                return cls.fromstring(value.value)
            case _:
                return cls.fromstring(value)

    @classmethod
    def fromtimedelta(cls, td: datetime.timedelta) -> 'DayTimeDuration':
        return cls(seconds=Decimal(
            '{}.{:06}'.format(td.days * 86400 + td.seconds, td.microseconds)
        ))

    def __init__(self, seconds: Union[Decimal, int] = 0) -> None:
        """
        :param seconds: a decimal or an integer instance that represents \
        days, hours, minutes, seconds and fractions of seconds.
        """
        super().__init__(0, seconds)

    def get_timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(
            seconds=int(self.seconds), microseconds=int(self.seconds % 1 * 1000000)
        )

    def __repr__(self) -> str:
        return '%s(seconds=%s)' % (self.__class__.__name__, normalized_seconds(self.seconds))

    def __add__(self, other: object) -> Union['DayTimeDuration', Time, AbstractDateTime]:
        if isinstance(other, (Time, Date)):
            return other + self
        elif isinstance(other, self.__class__):
            return DayTimeDuration(self.seconds + other.seconds)
        raise TypeError("cannot add %r to %r" % (type(other), type(self)))

    def __sub__(self, other: object) -> 'DayTimeDuration':
        if not isinstance(other, self.__class__):
            raise TypeError("cannot subtract %r from %r" % (type(other), type(self)))
        return DayTimeDuration(seconds=self.seconds - other.seconds)

    def __mul__(self, other: object) -> 'DayTimeDuration':
        if isinstance(other, (float, int, Decimal)):
            if math.isnan(other):
                raise ValueError("cannot multiply a %r by NaN" % type(self))

            if isinstance(other, (int, Decimal)):
                seconds = self.seconds * other
            else:
                seconds = self.seconds * Decimal.from_float(other)

            return DayTimeDuration(seconds)
        else:
            raise TypeError("cannot multiply a %r by %r" % (type(self), type(other)))

    def __truediv__(self, other: object) -> Union[Decimal, 'DayTimeDuration']:
        if isinstance(other, self.__class__):
            return self.seconds / other.seconds
        elif isinstance(other, (float, int, Decimal)):
            if math.isnan(other):
                raise ValueError("cannot divide a %r by NaN" % type(self))

            if isinstance(other, (int, Decimal)):
                seconds = self.seconds / other
            else:
                seconds = self.seconds / Decimal.from_float(other)

            return DayTimeDuration(seconds)
        else:
            raise TypeError("cannot divide a %r by %r" % (type(self), type(other)))



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import decimal
import math
import urllib.parse
from collections.abc import Iterator, Callable
from copy import copy
from decimal import Decimal
from itertools import product
from typing import Any, cast, ClassVar, SupportsFloat, TYPE_CHECKING, TypeVar

import elementpath.aliases as ta

from elementpath.exceptions import xpath_error, ElementPathError, ElementPathValueError, \
    ElementPathTypeError, MissingContextError
from elementpath.namespaces import XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE
from elementpath.namespaces import XSD_NAMESPACE, XPATH_MATH_FUNCTIONS_NAMESPACE
from elementpath.datatypes import AnyAtomicType, AbstractDateTime, AnyURI, \
    DayTimeDuration, Date, DateTime, DecimalProxy, Duration, Integer, QName, \
    Timezone, UntypedAtomic, AbstractQName
from elementpath.tdop import Token, MultiLabel
from elementpath.helpers import ordinal, get_double
from elementpath.xpath_context import XPathContext, XPathSchemaContext
from elementpath.xpath_nodes import XPathNode, NamespaceNode, DocumentNode, ElementNode
from elementpath.sequences import xlist

if TYPE_CHECKING:
    from . import XPathFunction, XPathArray, XPathConstructor  # noqa: F401
    from . import NameToken, TokenRegistry  # noqa: F401

_XSD_SPECIAL_TYPES = frozenset((XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE))
_CHILD_AXIS_TOKENS = frozenset((
    '*', 'node', 'child', 'text', '(name)', ':', '[', 'document-node',
    'element', 'comment', 'processing-instruction', 'schema-element'
))
_LEAF_ELEMENTS_TOKENS = frozenset((
    '(name)', '*', ':', '..', '.', '[', 'self', 'child', 'parent',
    'following-sibling', 'preceding-sibling', 'ancestor', 'ancestor-or-self',
    'descendant', 'descendant-or-self', 'following', 'preceding'
))

T = TypeVar('T', bound=ta.ItemType)


class XPathToken(Token[ta.XPathTokenType]):
    """Base class for XPath tokens."""
    registry: ClassVar['TokenRegistry']
    as_name: Callable[[], 'NameToken']

    parser: ta.XPathParserType
    value: ta.ValueType

    name: str = ''  # for storing the qualified name of a function
    namespace: str | None = None  # for namespace binding of names and wildcards
    occurrence: str = ''  # occurrence indicator, used by item types
    concatenated = False  # a flag for infix operators that can be concatenated

    def __str__(self) -> str:
        if self.symbol == '$':
            return '$%s variable reference' % (self[0].value if self._items else '')
        elif self.symbol == ',':
            return 'comma operator' if self.parser.version != '1.0' else 'comma symbol'
        elif self.symbol == '(':
            if not self or self[0].span[0] >= self.span[0]:
                return 'parenthesized expression'
            else:
                return 'function call expression'
        return super(XPathToken, self).__str__()

    @property
    def source(self) -> str:
        symbol = self.symbol
        if self.label == 'axis':
            # For XPath 2.0 'attribute' multirole token ('kind test', 'axis')
            return '%s::%s' % (symbol, self[0].source)
        elif symbol == '/' or symbol == '//':
            if not self:
                return symbol
            elif len(self) == 1:
                return f'{symbol}{self[0].source}'
            else:
                return f'{self[0].source}{symbol}{self[1].source}'
        elif symbol == '(':
            if not self:
                return '()'
            elif len(self) == 2:
                return f'{self[0].source}({self[1].source})'
            elif self[0].span[0] < self.span[0]:
                return f'{self[0].source}()'
            else:
                return f'({self[0].source})'
        elif symbol == '[':
            return '%s[%s]' % (self[0].source, self[1].source)
        elif symbol == ',':
            return '%s, %s' % (self[0].source, self[1].source)
        elif symbol == '$' or symbol == '@':
            return f'{symbol}{self[0].source}'
        elif symbol == '#':
            return '%s#%s' % (self[0].source, self[1].source)
        elif symbol == '{' or symbol == 'Q{':
            return '%s%s}%s' % (symbol, self[0].value, self[1].source)
        elif symbol == '=>':
            if isinstance(self[1], self.registry.function_token):
                return '%s => %s%s' % (self[0].source, self[1].symbol, self[2].source)
            return '%s => %s%s' % (self[0].source, self[1].source, self[2].source)
        elif symbol == 'if':
            return 'if (%s) then %s else %s' % (self[0].source, self[1].source, self[2].source)
        elif symbol == 'instance':
            return '%s instance of %s' % (
                self[0].source, ''.join(t.source for t in self[1:])
            )
        elif symbol in ('treat', 'cast', 'castable'):
            return '%s %s as %s' % (
                self[0].source, symbol, ''.join(t.source for t in self[1:])
            )
        elif symbol == 'for':
            return 'for %s return %s' % (
                ', '.join('%s in %s' % (self[k].source, self[k + 1].source)
                          for k in range(0, len(self) - 1, 2)),
                self[-1].source
            )
        elif symbol in ('every', 'some'):
            return '%s %s satisfies %s' % (
                symbol,
                ', '.join('%s in %s' % (self[k].source, self[k + 1].source)
                          for k in range(0, len(self) - 1, 2)),
                self[-1].source
            )
        elif symbol == 'let':
            return 'let %s return %s' % (
                ', '.join('%s := %s' % (self[k].source, self[k + 1].source)
                          for k in range(0, len(self) - 1, 2)),
                self[-1].source
            )
        elif symbol in ('-', '+') and len(self) == 1:
            return symbol + self[0].source
        return super(XPathToken, self).source

    @property
    def child_axis(self) -> bool:
        """Is `True` if the token apply child axis for default, `False` otherwise."""
        if self.symbol not in _CHILD_AXIS_TOKENS:
            return False
        elif self.symbol == '[':
            return self._items[0].child_axis
        elif self.symbol != ':':
            return True
        return not self._items[1].label.endswith('function')

    ###
    # Tokens tree analysis methods
    def iter_leaf_elements(self) -> Iterator[str]:
        """
        Iterates through the leaf elements of the token tree if there are any,
        returning QNames in prefixed format. A leaf element is an element
        positioned at last path step. Does not consider kind tests and wildcards.
        """
        if self.symbol in ('(name)', ':'):
            yield cast(str, self.value)
        elif self.symbol in ('//', '/'):
            if self._items[-1].symbol in _LEAF_ELEMENTS_TOKENS:
                yield from self._items[-1].iter_leaf_elements()

        elif self.symbol in ('[',):
            yield from self._items[0].iter_leaf_elements()
        else:
            for tk in self._items:
                yield from tk.iter_leaf_elements()

    ###
    # Evaluation and selectors methods
    def evaluate(self, context: ta.ContextType = None) -> ta.ValueType:
        """
        Evaluation method for XPath tokens.

        :param context: The XPath dynamic context.
        """
        return xlist(self.select(context))

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        """
        Select operator that generates XPath results, expanding sequences.

        :param context: The XPath dynamic context.
        """
        item = self.evaluate(context)
        if isinstance(item, list):
            yield from item
        else:
            yield item

    def select_flatten(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        """A select that flattens XPath results, including arrays."""
        for item in self.select(context):
            if isinstance(item, self.registry.array_token):
                yield from item.iter_flatten(context)
            else:
                yield item

    def select_with_focus(self, context: XPathContext) -> Iterator[ta.ItemType]:
        """Select item with an inner focus on dynamic context."""
        status = context.item, context.size, context.position, context.axis
        context.axis = None
        results = [x for x in self.select(context)]

        context.axis = None
        context.size = len(results)
        for context.position, context.item in enumerate(results, start=1):
            yield context.item

        context.item, context.size, context.position, context.axis = status

    def select_results(self, context: ta.ContextType) -> Iterator[ta.ResultType]:
        """
        Generates formatted XPath results.

        :param context: the XPath dynamic context.
        """
        if context is None:
            yield from self.select(context)
        else:
            self.parser.check_variables(context.variables)

            for result in self.select_flatten(context):
                if not isinstance(result, XPathNode):
                    yield result
                elif isinstance(result, NamespaceNode):
                    if self.parser.compatibility_mode:
                        yield result.prefix, result.uri
                    else:
                        yield result.uri
                elif isinstance(result, DocumentNode):
                    if result.is_extended:
                        # cannot represent with an ElementTree: yield the document node
                        yield result
                    elif result is context.root or result is not context.document:
                        yield result.value
                else:
                    yield result.value

    def get_results(self, context: ta.ContextType) -> \
            'list[ta.ResultType] | ta.AtomicType | XPathFunction':
        """
        Returns results formatted according to XPath specifications.

        :param context: the XPath dynamic context.
        :return: a list of nodes, atomic items or functions or a single atomic item or function.
        """
        results = list(self.select_results(context))
        if len(results) == 1:
            if hasattr(results[0], 'tag') or hasattr(results[0], 'getroot'):
                return results
            if isinstance(results[0], self.registry.function_token):
                return results[0]
            if self.symbol in ('.', '/', '//', '[', '(', '@', 'for',):
                return results
            if self.symbol in ('cast', 'castable', '-', '+', 'some', 'every'):
                return cast(AnyAtomicType, results[0])
            if self.label == 'kind test' or isinstance(self, self.registry.axis_token):
                return results
            if self.label == 'literal':
                return cast(AnyAtomicType, results[0])

            tk = self if self.symbol != ':' else self[1]
            if isinstance(tk, self.registry.function_token):
                rt = tk.sequence_types[-1]
                if rt.endswith(('*', '+')):
                    return results
                if rt.startswith('xs:'):
                    return cast(AnyAtomicType, results[0])
                return results

            if isinstance(results[0], AnyAtomicType):
                return results[0]
        return results

    ###
    # Helpers to get and validate arguments
    def get_argument(self, context: ta.ContextType,
                     index: int = 0,
                     required: bool = False,
                     default_to_context: bool = False,
                     default: ta.AtomicType | None = None,
                     cls: type[Any] | None = None,
                     promote: ta.ClassCheckType | None = None) -> Any:
        """
        Get the argument value of a function of constructor token. A zero length sequence is
        converted to a `None` value. If the function has no argument returns the context's
        item if the dynamic context is not `None`.

        :param context: the dynamic context.
        :param index: an index for select the argument to be got, the first for default.
        :param required: if set to `True` missing or empty sequence arguments are not allowed.
        :param default_to_context: if set to `True` then the item of the dynamic context is \
        returned when the argument is missing.
        :param default: the default value returned in case the argument is an empty sequence. \
        If not provided returns `None`.
        :param cls: if a type is provided performs a type checking on item.
        :param promote: a class or a tuple of classes that are promoted to `cls` class.
        """
        item: ta.ItemType | None
        try:
            token = self._items[index]
        except IndexError:
            if default_to_context:
                if context is None:
                    raise self.missing_context() from None
                item = context.item if context.item is not None else context.root
            elif isinstance(context, XPathSchemaContext):
                return default
            elif required:
                msg = "missing %s argument" % ordinal(index + 1)
                raise self.error('XPST0017', msg) from None
            else:
                return default
        else:
            if isinstance(token, XPathToken) and callable(token) and token.is_reference():
                return token  # It's a function reference

            item = None
            for k, result in enumerate(token.select(copy(context))):
                if k == 0:
                    item = result
                elif self.parser.compatibility_mode:
                    break
                elif isinstance(context, XPathSchemaContext):
                    # Multiple schema nodes are ignored but do not raise. The target
                    # of schema context selection is XSD type association and multiple
                    # node coherency is already checked at schema level.
                    break
                else:
                    msg = "a sequence of more than one item is not allowed as argument"
                    raise self.error('XPTY0004', msg)
            else:
                if item is None:
                    if not required or isinstance(context, XPathSchemaContext):
                        return default
                    ord_arg = ordinal(index + 1)
                    msg = "A not empty sequence required for {} argument"
                    raise self.error('XPTY0004', msg.format(ord_arg))

        if cls is not None:
            return self.validated_value(item, cls, promote, index)
        return item

    def get_argument_tokens(self) -> list['XPathToken']:
        """
        Builds and returns the argument tokens list, expanding the comma tokens.
        """
        tk = self
        tokens = []
        while True:
            if tk.symbol == ',':
                tokens.append(tk[1])
                tk = tk[0]
            else:
                tokens.append(tk)
                return tokens[::-1]

    def get_function(self, context: ta.ContextType, arity: int = 0) -> 'XPathFunction':
        if isinstance(self, self.registry.function_token):
            func = self
        elif self.symbol in (':', 'Q{') and isinstance(self[1], self.registry.function_token):
            func = self[1]
        elif self.symbol == '(name)':
            msg = f'unknown function: {self.value}#{arity}'
            raise self.error('XPST0017', msg)
        else:
            item = self.evaluate(context)
            if not isinstance(item, self.registry.function_token):
                msg = f'unknown function: {item}#{arity}'
                raise self.error('XPST0017', msg)
            func = item

        max_args = func.max_args
        if func.min_args > arity or max_args is not None and max_args < arity:
            msg = f'unknown function: {func.symbol}#{arity}'
            raise self.error('XPST0017', msg)

        return func

    def validated_value(self, item: Any, cls: type[Any],
                        promote: ta.ClassCheckType | None = None,
                        index: int | None = None) -> Any:
        """
        type promotion checking (see "function conversion rules" in XPath 2.0 language definition)
        """
        if isinstance(item, cls) or isinstance(item, XPathToken) and item.name == '(value)':
            return item
        elif promote and isinstance(item, promote):
            return cls(item)

        if self.parser.compatibility_mode:
            if issubclass(cls, str):
                return self.string_value(item)
            elif issubclass(cls, float) or issubclass(float, cls):
                return self.number_value(item)

        if issubclass(cls, XPathToken) or self.parser.version == '1.0':
            code = 'XPTY0004'
        else:
            value = self.data_value(item)

            if isinstance(value, cls):
                return value
            elif isinstance(value, AnyURI) and issubclass(cls, str):
                return cls(value)
            elif isinstance(value, UntypedAtomic):
                try:
                    return cls(value)
                except (TypeError, ValueError):
                    pass

            if value == []:
                code = 'FOTY0012'
            else:
                code = 'XPTY0004'

        if index is None:
            msg = f"item type is {type(item)!r} instead of {cls!r}"
        else:
            msg = f"{ordinal(index+1)} argument has type {type(item)!r} instead of {cls!r}"
        raise self.error(code, msg)

    ###
    # Atomization of sequences
    def atomization(self, context: ta.ContextType = None) -> Iterator[ta.AtomicType]:
        """
        Helper method for value atomization of a sequence.

        Ref: https://www.w3.org/TR/xpath31/#id-atomization

        :param context: the XPath dynamic context.
        """
        for item in self.select(copy(context)):
            yield from self.atomize_item(item)

    def atomize_item(self, item: ta.ValueType) -> Iterator[ta.AtomicType]:
        """
        Atomization of a sequence item. Yields typed values, as computed by
        fn:data().

        Ref: https://www.w3.org/TR/xpath31/#id-atomization
             https://www.w3.org/TR/xpath20/#dt-typed-value
        """
        match item:
            case None:
                return
            case XPathNode():
                if self.parser.version != '1.0':
                    value = None
                    for value in item.iter_typed_values:
                        yield value

                    if value is None:
                        msg = f"argument node {item!r} does not have a typed value"
                        raise self.error('FOTY0012', msg)
                else:
                    value = item.compat_string_value
                    yield value

            case list():
                for v in item:
                    yield from self.atomize_item(v)

            case XPathToken():
                if not callable(item):
                    msg = f"sequence item {item!r} is not appropriate for the context"
                    raise self.error('XPTY0004', msg)
                elif item.label != 'array':
                    raise self.error('FOTY0013', f"{item.label!r} has no typed value")

                for v in cast('XPathArray', item).iter_flatten():
                    if isinstance(v, AnyAtomicType):
                        yield v
                    else:
                        yield from self.atomize_item(v)

            case AnyAtomicType():
                yield cast(ta.AtomicType, item)
            case bytes():
                yield item.decode()
            case _:
                msg = f"sequence item {item!r} is not appropriate for the context"
                raise self.error('XPTY0004', msg)

    ###
    # Helpers for operators
    def get_atomized_operand(self, context: ta.ContextType = None) -> ta.AtomicType | None:
        """
        Get the atomized value for an XPath operator.

        :param context: the XPath dynamic context.
        :return: the atomized value of a single length sequence or `None` if the sequence is empty.
        """
        value = None
        first = True
        for value in self.atomization(context):
            if not first:
                msg = "atomized operand is a sequence of length greater than one"
                raise self.error('XPTY0004', msg)
            first = False
        else:
            if isinstance(value, UntypedAtomic):
                return str(value)
            else:
                return value

    def iter_comparison_data(self, context: ta.ContextType) -> Iterator[Any]:
        """
        Generates comparison data couples for the general comparison of sequences.
        Different sequences maybe generated with an XPath 2.0 parser, depending on
        compatibility mode setting.

        Ref: https://www.w3.org/TR/xpath20/#id-general-comparisons

        :param context: the XPath dynamic context.
        """
        left_values: Any
        right_values: Any
        msg = "cannot compare {!r} and {!r}"

        if self.parser.compatibility_mode:
            left_values = [x for x in self._items[0].atomization(context)]
            right_values = [x for x in self._items[1].atomization(context)]
            # Boolean comparison if one of the results is a single boolean value (1.)
            try:
                if isinstance(left_values[0], bool):
                    if len(left_values) == 1:
                        yield left_values[0], self.boolean_value(right_values)
                        return
                if isinstance(right_values[0], bool):
                    if len(right_values) == 1:
                        yield self.boolean_value(left_values), right_values[0]
                        return
            except IndexError:
                return

            # Converts to float for lesser-greater operators (3.)
            if self.symbol in ('<', '<=', '>', '>='):
                yield from product(map(float, left_values), map(float, right_values))
                return
            elif self.parser.version == '1.0':
                yield from product(left_values, right_values)
                return
        else:
            left_values = self._items[0].atomization(context)
            right_values = self._items[1].atomization(context)

        for op1, op2 in product(left_values, right_values):
            match op1:
                case str() | AnyURI():
                    if not isinstance(op2, (str, UntypedAtomic, AnyURI)):
                        raise TypeError(msg.format(type(op1), type(op2)))
                case bool():
                    if isinstance(op2, (str, Integer, AbstractQName, AnyURI)):
                        raise TypeError(msg.format(type(op1), type(op2)))
                case Integer():
                    if isinstance(op2, (str, AbstractQName, AnyURI, bool)):
                        raise TypeError(msg.format(type(op1), type(op2)))
                case float():
                    if isinstance(op2, decimal.Decimal):
                        yield op1, float(op2)
                        continue
                    elif isinstance(op2, (str, AbstractQName, AnyURI, bool)):
                        raise TypeError(msg.format(type(op1), type(op2)))
                case decimal.Decimal():
                    if isinstance(op2, float):
                        yield float(op1), op2
                        continue
                    elif isinstance(op2, (str, AbstractQName, AnyURI, bool)):
                        raise TypeError(msg.format(type(op1), type(op2)))
                case AbstractQName():
                    if not isinstance(op2, (AbstractQName, UntypedAtomic)):
                        raise TypeError(msg.format(type(op1), type(op2)))

            yield op1, op2

    def get_operands(self, context: ta.ContextType, cls: type[Any] | None = None) -> Any:
        """
        Returns the operands for a binary operator. Float arguments are converted
        to decimal if the other argument is a `Decimal` instance.

        :param context: the XPath dynamic context.
        :param cls: if a type is provided performs a type checking on item.
        :return: a couple of values representing the operands. If any operand \
        is not available returns a `(None, None)` couple.
        """
        op1 = self.get_argument(context, cls=cls)
        if op1 is None:
            return None, None
        elif isinstance(op1, ElementNode):
            op1 = self._items[0].data_value(op1)

        op2 = self.get_argument(context, index=1, cls=cls)
        if op2 is None:
            return None, None
        elif isinstance(op2, ElementNode):
            op2 = self._items[1].data_value(op2)

        if isinstance(op1, AbstractDateTime) and isinstance(op2, AbstractDateTime):
            if context is not None and context.timezone is not None:
                if op1.tzinfo is None:
                    op1.tzinfo = context.timezone
                if op2.tzinfo is None:
                    op2.tzinfo = context.timezone
        else:
            if isinstance(op1, UntypedAtomic):
                op1 = self.cast_to_double(op1.value)
                if isinstance(op2, Decimal):
                    return op1, float(op2)
            if isinstance(op2, UntypedAtomic):
                op2 = self.cast_to_double(op2.value)
                if isinstance(op1, Decimal):
                    return float(op1), op2

        if isinstance(op1, float):
            if isinstance(op2, Duration):
                return Decimal(op1), op2
            if isinstance(op2, Decimal):
                return op1, type(op1)(op2)
        if isinstance(op2, float):
            if isinstance(op1, Duration):
                return op1, Decimal(op2)
            if isinstance(op1, Decimal):
                return type(op2)(op1), op2

        return op1, op2

    def get_absolute_uri(self, uri: str, base_uri: str | None = None) -> str:
        """
        Obtains an absolute URI from the argument and the static context.

        :param uri: a string representing a URI.
        :param base_uri: an alternative base URI, otherwise the base_uri \
        of the static context is used.
        :returns: the argument if it's an absolute URI, otherwise returns the URI
        obtained by the join o the base_uri of the static context with the
        argument. Returns the argument if the base_uri is `None`.
        """
        if not base_uri:
            base_uri = self.parser.base_uri

        uri_parts: urllib.parse.ParseResult = urllib.parse.urlparse(uri)
        if uri_parts.scheme or uri_parts.netloc or base_uri is None:
            return uri

        base_uri_parts: urllib.parse.SplitResult = urllib.parse.urlsplit(base_uri)
        if base_uri_parts.fragment or not base_uri_parts.scheme and \
                not base_uri_parts.netloc and not base_uri_parts.path.startswith('/'):
            raise self.error('FORG0002', '{!r} is not suitable as base URI'.format(base_uri))

        if uri_parts.path.startswith('/') and base_uri_parts.path not in ('', '/'):
            return uri
        return urllib.parse.urljoin(base_uri, uri)

    def bind_namespace(self, namespace: str) -> None:
        """
        Bind a token with a namespace. The token has to be a name, a name wildcard,
        a function or a constructor, otherwise a syntax error is raised. Functions
        and constructors must be limited to their namespaces.
        """
        if self.symbol in ('(name)', '*') or isinstance(self, self.registry.proxy_token):
            pass
        elif namespace == self.parser.function_namespace:
            if self.label != 'function' and self.label != 'external function':
                msg = "a name, a wildcard or a function expected"
                raise self.wrong_syntax(msg, code='XPST0017')
            elif isinstance(self.label, MultiLabel):
                self.label = 'function'
        elif namespace == XSD_NAMESPACE:
            if self.label != 'constructor function':
                msg = "a name, a wildcard or a constructor function expected"
                raise self.wrong_syntax(msg, code='XPST0017')
            elif isinstance(self.label, MultiLabel):
                self.label = 'constructor function'
        elif namespace == XPATH_MATH_FUNCTIONS_NAMESPACE:
            if self.label != 'math function':
                msg = "a name, a wildcard or a math function expected"
                raise self.wrong_syntax(msg, code='XPST0017')
            elif isinstance(self.label, MultiLabel):
                self.label = 'math function'
        elif not isinstance(self, self.registry.function_token):
            msg = "a name, a wildcard or a function expected"
            raise self.wrong_syntax(msg, code='XPST0017')
        elif self.namespace and namespace != self.namespace:
            msg = "unmatched namespace"
            raise self.wrong_syntax(msg, code='XPST0017')

        self.namespace = namespace
        self.name = f'{{{namespace}}}{self.value}'

    def adjust_datetime(self, context: ta.ContextType, cls: type[AbstractDateTime]) \
            -> ta.OneOrEmpty[AbstractDateTime | DayTimeDuration]:
        """
        XSD datetime adjust function helper.

        :param context: the XPath dynamic context.
        :param cls: the XSD datetime subclass to use.
        :return: an empty list if there is only one argument that is the empty sequence \
        or the adjusted XSD datetime instance.
        """
        timezone: Any | None
        item: AbstractDateTime | None
        _item: AbstractDateTime | DayTimeDuration

        if len(self) == 1:
            item = self.get_argument(context, cls=cls)
            if item is None:
                return []
            timezone = getattr(context, 'timezone', None)
        else:
            item = self.get_argument(context, cls=cls)
            timezone = self.get_argument(context, 1, cls=DayTimeDuration)

            if timezone is not None:
                try:
                    timezone = Timezone.fromduration(timezone)
                except ValueError as err:
                    if isinstance(context, XPathSchemaContext):
                        timezone = Timezone.fromduration(DayTimeDuration(0))
                    else:
                        raise self.error('FODT0003', str(err)) from None
            if item is None:
                return []

        _item = copy(item)
        _tzinfo = _item.tzinfo
        try:
            if isinstance(_tzinfo, Timezone) and isinstance(timezone, Timezone):
                if isinstance(_item, DateTime):
                    _item += timezone.offset
                elif not isinstance(item, Date):
                    _item += timezone.offset - _tzinfo.offset
                elif timezone.offset < _tzinfo.offset:
                    _item -= timezone.offset - _tzinfo.offset
                    _item -= DayTimeDuration.fromstring('P1D')
        except OverflowError as err:
            if isinstance(context, XPathSchemaContext):
                return _item
            raise self.error('FODT0001', str(err)) from None

        if isinstance(_item, AbstractDateTime):
            _item.tzinfo = timezone
        return _item

    ###
    # Helpers for cast to XSD types
    def cast_to_qname(self, value: str) -> QName:
        """Cast a string compatible value to a QName object."""
        try:
            return QName.make(value, parser=self.parser)
        except ValueError:
            msg = 'invalid value {!r} for an xs:QName'.format(value)
            raise self.error('FORG0001', msg) from None
        except KeyError as err:
            msg = 'no namespace found for prefix {}'.format(err)
            raise self.error('FONS0004', msg) from None
        except TypeError:
            msg = 'the argument has an invalid type {!r}'.format(type(value))
            raise self.error('XPTY0004', msg) from None

    def cast_to_double(self, value: SupportsFloat | str) -> float:
        """Cast a value to xs:double."""
        try:
            return get_double(value, self.parser.xsd_version)
        except ValueError as err:
            raise self.error('FORG0001', str(err))  # str or UntypedAtomic

    def cast_to_primitive_type(self, value: Any, type_name: str) -> Any:
        """Cast a value to a primitive atomic type or a list of atomic values."""
        if value is None or not type_name.startswith('xs:') or type_name.count(':') != 1:
            return value

        type_name = type_name[3:].rstrip('+*?')
        token = cast('XPathConstructor', self.parser.symbol_table[type_name](self.parser))

        def cast_value(v: Any) -> Any:
            try:
                if isinstance(v, (UntypedAtomic, AnyURI)):
                    return token.cast(v)
                elif isinstance(v, (float, DecimalProxy)):
                    if type_name in ('double', 'float'):
                        return token.cast(v)
            except (ValueError, TypeError):
                return v
            else:
                return v

        if isinstance(value, list):
            return xlist([cast_value(x) for x in value])
        else:
            return cast_value(value)

    ###
    # XPath data accessors base functions
    def boolean_value(self, obj: Any) -> bool:
        """
        The effective boolean value, as computed by fn:boolean().
        """
        if isinstance(obj, list):
            if not obj:
                return False
            elif isinstance(obj[0], XPathNode):
                return True
            elif len(obj) > 1:
                message = "effective boolean value is not defined for a sequence " \
                          "of two or more items not starting with an XPath node."
                raise self.error('FORG0006', message)
            else:
                obj = obj[0]
        elif isinstance(obj, Iterator):
            k = 0
            items = obj
            obj = None
            for obj in items:
                if k:
                    message = "effective boolean value is not defined for a sequence " \
                              "of two or more items not starting with an XPath node."
                    raise self.error('FORG0006', message)
                elif isinstance(obj, XPathNode):
                    return True
                k += 1
            else:
                if not k:
                    return False

        if isinstance(obj, (int, str, UntypedAtomic, AnyURI)):  # Include bool
            return bool(obj)
        elif isinstance(obj, (float, Decimal)):
            return False if math.isnan(obj) else bool(obj)
        elif obj is None:
            return False
        elif isinstance(obj, XPathNode):
            return True
        else:
            message = "effective boolean value is not defined for {!r}.".format(type(obj))
            raise self.error('FORG0006', message)

    def data_value(self, obj: Any) -> ta.AtomicType | None:
        """
        Returns the typed value. Raises an error if the atomization of the value
        produces more than one typed value.
        """
        value = None
        first = True
        for value in self.atomize_item(obj):
            if not first:
                msg = "atomized value is a sequence of length greater than one"
                raise self.error('XPTY0004', msg)
            first = False
        else:
            return value

    def string_value(self, obj: Any) -> str:
        """
        The string value, as computed by fn:string().
        """
        if obj is None:
            return ''
        elif isinstance(obj, XPathNode):
            if self.parser.version == '1.0':
                return obj.compat_string_value
            return obj.string_value
        elif isinstance(obj, bool):
            return 'true' if obj else 'false'
        elif isinstance(obj, Decimal):
            value = format(obj, 'f')
            if '.' in value:
                return value.rstrip('0').rstrip('.')
            return value

        elif isinstance(obj, float):
            if math.isnan(obj):
                return 'NaN'
            elif math.isinf(obj):
                return str(obj).upper()

            value = str(obj)
            if '.' in value:
                value = value.rstrip('0').rstrip('.')
            if '+' in value:
                value = value.replace('+', '')
            if 'e' in value:
                return value.upper()
            return value

        elif isinstance(obj, self.registry.function_token):
            if self.symbol in ('concat', '||'):
                raise self.error('FOTY0013', f"an argument of {self} is a function")
            else:
                raise self.error('FOTY0014', f"{obj.label!r} has no string value")

        return str(obj)

    def number_value(self, obj: Any) -> float:
        """
        The numeric value, as computed by fn:number() on each item. Returns a float value.
        """
        try:
            if isinstance(obj, XPathNode):
                if self.parser.version == '1.0':
                    return get_double(obj.compat_string_value, self.parser.xsd_version)
                return get_double(obj.string_value, self.parser.xsd_version)
            else:
                return get_double(obj, self.parser.xsd_version)
        except (TypeError, ValueError):
            return math.nan

    ###
    # Error handling helpers and shortcuts
    def error(self, code: str | QName,
              message_or_error: str | Exception | None = None) -> ElementPathError:
        return xpath_error(code, message_or_error, self, self.parser.namespaces)

    def expected(self, *symbols: str,
                 message: str | None = None,
                 code: str = 'XPST0003') -> None:
        if symbols and self.symbol not in symbols:
            raise self.wrong_syntax(message, code)

    def unexpected(self, *symbols: str,
                   message: str | None = None,
                   code: str = 'XPST0003') -> None:
        if not symbols or self.symbol in symbols:
            raise self.wrong_syntax(message, code)

    def wrong_syntax(self, message: str | None = None,  # type: ignore[override]
                     code: str = 'XPST0003') -> ElementPathError:
        if self.label == 'function':
            code = 'XPST0017'

        if message:
            return self.error(code, message)

        error = super(XPathToken, self).wrong_syntax(message)
        return self.error(code, str(error))

    def wrong_value(self, message: str | None = None) -> ElementPathValueError:
        return cast(ElementPathValueError, self.error('FOCA0002', message))

    def wrong_type(self, message: str | None = None) -> ElementPathTypeError:
        return cast(ElementPathTypeError, self.error('FORG0006', message))

    def missing_context(self, message: str | None = None) -> MissingContextError:
        return cast(MissingContextError, self.error('XPDY0002', message))



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import re
import math
import operator
from calendar import isleap, leapdays
from collections.abc import Mapping
from decimal import Decimal
from typing import Any, Generic, Optional, SupportsFloat, SupportsIndex, TypeVar, Union
from urllib.parse import urlsplit

###
# Common sets constants
OCCURRENCE_INDICATORS = frozenset(('?', '*', '+'))
BOOLEAN_VALUES = frozenset(('true', 'false', '1', '0'))
NUMERIC_INF_OR_NAN = frozenset(('INF', '-INF', '+INF', 'NaN'))
INVALID_NUMERIC = frozenset(
    ('inf', '+inf', '-inf', 'nan', 'infinity', '+infinity', '-infinity')
)

MathArgType = Union[SupportsFloat, SupportsIndex]
FloatArgType = Union[SupportsFloat, SupportsIndex, str]

T = TypeVar('T')


class Property(Generic[T]):
    """A descriptor for managing protected class properties."""
    __slots__ = ('_name', '_value')

    def __init__(self, value: T) -> None:
        self._value = value

    def __set_name__(self, owner: type[Any], name: str) -> None:
        self._name = name

    def __get__(self, instance: Any, owner: type[Any]) -> T:
        return self._value

    def __set__(self, instance: Any, value: Any) -> None:
        raise AttributeError("Can't set attribute {}".format(self._name))

    def __delete__(self, instance: Any) -> None:
        raise AttributeError("Can't delete attribute {}".format(self._name))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._value!r})'

    @property
    def value(self) -> T:
        return self._value


###
# Data validation patterns

class LazyPattern:
    """
    A descriptor for creating lazy regexp patterns. The compiled pattern is built
    only when the descriptor attribute is accessed (e.g. a hasattr() call).
    """
    _compiled: re.Pattern[str]

    __slots__ = ('_name', '_pattern', '_flags', '_compiled')

    def __init__(self, pattern: str, flags: Union[int, re.RegexFlag] = 0) -> None:
        self._pattern = pattern
        self._flags = flags

    def __set_name__(self, owner: type[Any], name: str) -> None:
        self._name = name

    def __get__(self, instance: Optional[Any], owner: type[Any]) -> re.Pattern[str]:
        try:
            return self._compiled
        except AttributeError:
            self._compiled = re.compile(self._pattern, self._flags)
            return self._compiled

    def __set__(self, instance: Any, value: Any) -> None:
        raise AttributeError("Can't set attribute {}".format(self._name))

    def __delete__(self, instance: Any) -> None:
        raise AttributeError("Can't delete attribute {}".format(self._name))

    @property
    def groupindex(self) -> Mapping[str, int] | None:
        try:
            return self._compiled.groupindex
        except AttributeError:
            self._compiled = re.compile(self._pattern, self._flags)
            return self._compiled.groupindex

    def match(self, string: str) -> re.Match[str] | None:
        try:
            return self._compiled.match(string)
        except AttributeError:
            self._compiled = re.compile(self._pattern, self._flags)
            return self._compiled.match(string)

    def search(self, string: str) -> re.Match[str] | None:
        try:
            return self._compiled.search(string)
        except AttributeError:
            self._compiled = re.compile(self._pattern, self._flags)
            return self._compiled.search(string)


class Patterns:
    """
    Helper patterns, the ones that aren't used at import time are defined lazy.
    """
    whitespaces = re.compile(r'[^\S\xa0]+')  # include ASCII 160 (non-breaking space)
    normalize = LazyPattern(r'[^\S\xa0]')
    ncname = LazyPattern(r'^[^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*$')
    extended_qname = LazyPattern(
        r'^(?:Q{(?P<namespace>[^}]+)}|'
        r'(?P<prefix>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*):)?'
        r'(?P<local>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*)$',
    )
    replacement = LazyPattern(r'^([^\\$]|\\{2}|\\\$|\$\d+)*$')
    sequence_type = LazyPattern(r'\s?([()?*+,])\s?')
    unicode_escape = LazyPattern(r'(?:\\u([0-9A-Fa-f]{4})|\\U([0-9A-Fa-f]{8}))')
    wrong_escape = LazyPattern(r'%(?![a-fA-F\d]{2})')
    xml_newlines = LazyPattern('\r\n|\r|\n')

    # Regex patterns related to names and namespaces
    namespace_uri = LazyPattern(r'{([^}]+)}')
    expanded_name = LazyPattern(
        r'^(?:(?:Q{|{)(?P<namespace>[^}]*)})?'
        r'(?P<local>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*)$',
    )
    unbound_expanded_name = LazyPattern(
        r'(?:Q{|{)([^}]*)}[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*'
    )
    unbound_qname = LazyPattern(
        r'(?:(?P<prefix>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*):)?'
        r'(?P<local>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*)',
    )


def upper_camel_case(s: str) -> str:
    return re.sub(r'^\d+', '', re.sub(r'[\W_]', '', s.title()))


def collapse_white_spaces(s: str) -> str:
    return Patterns.whitespaces.sub(' ', s).strip(' ')


def is_ncname(s: str) -> bool:
    return Patterns.ncname.match(s) is not None


def is_idrefs(value: Optional[str]) -> bool:
    return isinstance(value, str) and \
        all(Patterns.ncname.match(x) is not None for x in value.split())


###
# Operators
node_position = operator.attrgetter('position')


def reversed_sub(a: Any, b: Any) -> Any:
    return operator.sub(b, a)


def reversed_truediv(a: Any, b: Any) -> Any:
    return operator.truediv(b, a)


###
# Date/Time helpers
MONTH_DAYS = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
MONTH_DAYS_LEAP = (0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)


def adjust_day(year: int, month: int, day: int) -> int:
    match month:
        case 1 | 3 | 5 | 7 | 8 | 10 | 12:
            return day
        case 4 | 6 | 9 | 11:
            return min(day, 30)
        case _:
            return min(day, 29) if isleap(year) else min(day, 28)


def days_from_common_era(year: int) -> int:
    """
    Returns the number of days from 0001-01-01 to the provided year. For a
    common era year the days are counted until the last day of December, for a
    BCE year the days are counted down from the end to the 1st of January.
    """
    match year:
        case year if year > 0:
            return year * 365 + year // 4 - year // 100 + year // 400
        case year if year >= -1:
            return year * 366
        case _:
            year = -year - 1
            return -(366 + year * 365 + year // 4 - year // 100 + year // 400)


DAYS_IN_4Y = days_from_common_era(4)
DAYS_IN_100Y = days_from_common_era(100)
DAYS_IN_400Y = days_from_common_era(400)


def months2days(year: int, month: int, months_delta: int) -> int:
    """
    Converts a delta of months to a delta of days, counting from the 1st day of the month,
    relative to the year and the month passed as arguments.

    :param year: the reference start year, a negative or zero value means a BCE year \
    (0 is 1 BCE, -1 is 2 BCE, -2 is 3 BCE, etc.).
    :param month: the starting month (1-12).
    :param months_delta: the number of months, if negative count backwards.
    """
    if not months_delta:
        return 0

    total_months = month - 1 + months_delta
    target_year = year + total_months // 12
    target_month = total_months % 12 + 1

    if month <= 2:
        y_days = 365 * (target_year - year) + leapdays(year, target_year)
    else:
        y_days = 365 * (target_year - year) + leapdays(year + 1, target_year + 1)

    months_days = MONTH_DAYS_LEAP if isleap(target_year) else MONTH_DAYS
    if target_month >= month:
        m_days = sum(months_days[m] for m in range(month, target_month))
        return y_days + m_days if y_days >= 0 else y_days + m_days
    else:
        m_days = sum(months_days[m] for m in range(target_month, month))
        return y_days - m_days if y_days >= 0 else y_days - m_days


def round_number(value: Union[float, int, Decimal]) -> Union[float, int, Decimal]:
    if math.isnan(value) or math.isinf(value):
        return value

    number = Decimal(value)
    if number > 0:
        return type(value)(number.quantize(Decimal('1'), rounding='ROUND_HALF_UP'))
    else:
        return type(value)(number.quantize(Decimal('1'), rounding='ROUND_HALF_DOWN'))


def normalized_seconds(seconds: Union[int, Decimal]) -> str:
    # Decimal.normalize() does not remove exp every time: eg. Decimal('1E+1')
    return '{:.6f}'.format(seconds).rstrip('0').rstrip('.')


def is_xml_codepoint(cp: int) -> bool:
    return cp in (0x9, 0xA, 0xD) or \
        0x20 <= cp <= 0xD7FF or \
        0xE000 <= cp <= 0xFFFD or \
        0x10000 <= cp <= 0x10FFFF


def ordinal(n: int) -> str:
    match n:
        case 11 | 12 | 13:
            return '%dth' % n
        case n if n % 10 == 1:
            return '%dst' % n
        case n if n % 10 == 2:
            return '%dnd' % n
        case n if n % 10 == 3:
            return '%drd' % n
        case _:
            return '%dth' % n


def get_double(value: FloatArgType, xsd_version: str | None = None) -> float:
    if isinstance(value, str):
        value = collapse_white_spaces(value)
        if value in NUMERIC_INF_OR_NAN and (xsd_version != '1.0' or value != '+INF'):
            if value == 'NaN':
                return math.nan  # for NaN use the predefined instance to keep identity
        elif value.lower() in INVALID_NUMERIC:
            raise ValueError(f'invalid value {value!r} for xs:double/xs:float')
    elif math.isnan(value):
        return math.nan

    return float(value)


def numeric_equal(op1: MathArgType, op2: MathArgType) -> bool:
    if op1 == op2:
        return True
    return math.isclose(op1, op2, rel_tol=1e-7, abs_tol=0.0)


def numeric_not_equal(op1: MathArgType, op2: MathArgType) -> bool:
    if op1 == op2:
        return False
    return not math.isclose(op1, op2, rel_tol=1e-7, abs_tol=0.0)


def equal(op1: Any, op2: Any) -> bool:
    if isinstance(op1, float) and math.isnan(op1):
        return isinstance(op2, float) and math.isnan(op2)
    return bool(op1 == op2)


def not_equal(op1: Any, op2: Any) -> bool:
    if isinstance(op1, float) and math.isnan(op1):
        return not isinstance(op2, float) or not math.isnan(op2)
    return bool(op1 != op2)


def match_wildcard(name: Optional[str], wildcard: str) -> bool:
    if not name:
        return False
    elif wildcard in ('*', '{*}*'):
        return True
    elif wildcard == '{}*':
        return name[0] != '{' or name[:2] == '{}'
    elif wildcard[-1] == '*':
        return name.startswith(wildcard[:-1])
    elif not wildcard.startswith('{*}'):
        return False
    elif name[0] == '{':
        return name.endswith(wildcard[2:])
    else:
        return name == wildcard[3:]


def escape_json_string(s: str, escaped: bool = False) -> str:
    if escaped:
        s = s.replace('\\"', '"')
    else:
        s = s.replace('\\', '\\\\')

    s = s.replace('\"', '\\"').\
        replace('\b', r'\b').\
        replace('\r', r'\r').\
        replace('\n', r'\n').\
        replace('\t', r'\t').\
        replace('\f', r'\f').\
        replace('/', r'\/')
    return ''.join(
        rf'\u{ord(x):04X}' if 1 <= ord(x) <= 31 or 127 <= ord(x) <= 159 else x
        for x in s
    )


def unescape_json_string(s: str) -> str:

    def unicode_escape_callback(match: re.Match[str]) -> str:
        group = match.group(1) or match.group(2)
        return chr(int(group.upper(), 16))

    s = s.replace('\\"', '\"').\
        replace(r'\b', '\b').\
        replace(r'\r', '\r').\
        replace(r'\n', '\n').\
        replace(r'\t', '\t').\
        replace(r'\f', '\f').\
        replace(r'\/', '/').\
        replace('\\\\', '\\')

    return Patterns.unicode_escape.sub(unicode_escape_callback, s)


def split_function_test(function_test: str) -> list[str]:
    if not function_test.startswith('function('):
        return []
    elif function_test == 'function(*)':
        return ['*']

    parts = function_test[9:].partition(') as ')
    if parts[0]:
        sequence_types = parts[0].split(', ')
        sequence_types.append(parts[2])
    else:
        sequence_types = [parts[2]]

    return sequence_types


def is_absolute_uri(uri: str) -> bool:
    try:
        parts = urlsplit(uri.strip())
    except ValueError:
        return False
    else:
        return parts.scheme == 'urn' or \
            parts.scheme != '' and parts.netloc != '' or \
            parts.path.startswith('/')



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 3.1 implementation - part 2 (operators and constructors)
"""
from collections.abc import Iterator, Iterable
from typing import cast, Union

import elementpath.aliases as ta

from elementpath.sequence_types import is_sequence_type, match_sequence_type
from elementpath.xpath_tokens import XPathToken, ProxyToken, XPathFunction, \
    XPathMap, XPathArray
from elementpath.sequences import xlist

from .xpath31_parser import XPath31Parser

__all__ = ['XPath31Parser']

register = XPath31Parser.register
method = XPath31Parser.method
function = XPath31Parser.function

register('map', bp=90, label=('kind test', 'map'), bases=(XPathFunction,),
         pattern=r'(?<!\$)\bmap(?=\s*(?:\(\:.*\:\))?\s*(?=\(|\{)(?!\:))')


@method('map')
def nud__map_sequence_type_or_constructor(self: XPathFunction) \
        -> Union[XPathToken, XPathMap, XPathArray]:
    if self.parser.next_token.symbol == '{':
        self.parser.token = XPathMap(self.parser).nud()
        return self.parser.token
    elif self.parser.next_token.symbol != '(':
        return self.as_name()

    self.label = 'kind test'

    self.parser.advance('(')
    if self.parser.next_token.label not in ('kind test', 'sequence type', 'function test'):
        self.parser.expected_next('(name)', ':', '*', message='a QName or a wildcard expected')
    self[:] = self.parser.expression(45),
    self.parser.parse_occurrence(self[0])

    if self[0].symbol != '*':
        self.parser.advance(',')
        if self.parser.next_token.label not in ('kind test', 'sequence type', 'function test'):
            self.parser.expected_next('(name)', ':', '*', message='a QName or a wildcard expected')
        self.append(self.parser.expression(45))
        self.parser.parse_occurrence(self[-1])

    self.parser.advance(')')
    return self


register('array', bp=90, label=('kind test', 'array'), bases=(XPathFunction,),
         pattern=r'(?<!\$)\barray(?=\s*(?:\(\:.*\:\))?\s*(?=\(|\{)(?!\:))')


@method('array')
def nud__sequence_type_or_curly_array_constructor(self: XPathFunction) -> XPathToken:
    if self.parser.next_token.symbol == '{':
        self.parser.token = XPathArray(self.parser).nud()
        return self.parser.token
    elif self.parser.next_token.symbol != '(':
        return self.as_name()

    self.label = 'kind test'
    self.parser.advance('(')
    if self.parser.next_token.label not in ('kind test', 'function test'):
        self.parser.expected_next('(name)', ':', '*', 'item')
    self[:] = self.parser.expression(45),
    if self[0].symbol != '*':
        self.parser.parse_occurrence(self[0])
    self.parser.advance(')')
    self.parser.parse_occurrence(self)
    return self


@method('map')
@method('array')
def select__map_or_array_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[Union[XPathMap, XPathArray]]:
    if context is None:
        raise self.missing_context()

    for item in context.iter_children_or_self():
        if match_sequence_type(item, self.source, self.parser):
            yield cast(Union[XPathMap, XPathArray], item)


###
# Square array constructor (pushed lazy)
@method('[')
def nud__square_array_constructor(self: XPathToken) -> XPathToken:
    if self.parser.version < '3.1':
        raise self.wrong_syntax()

    # Constructs an XPathArray token and returns it instead of the predicate
    token = XPathArray(self.parser)
    token.symbol = '['
    if token.parser.next_token.symbol not in (']', '(end)'):
        while True:
            token.append(self.parser.expression(5))
            if token.parser.next_token.symbol != ',':
                break
            token.parser.advance()

    token.parser.advance(']')
    return token


class LookupOperatorToken(XPathToken):
    """
    Question mark symbol is used for XP31+ lookup operator and also for
    placeholder in XP30+ partial functions and for optional occurrences.
    """
    symbol = lookup_name = '?'
    lbp = 85
    rbp = 85

    def __init__(self, parser: ta.XPathParserType, value: ta.AtomicType | None = None) -> None:
        super().__init__(parser, value)
        if self.parser.token.symbol in ('(', ','):
            # It's a placeholder symbol or a unary lookup operator
            # in a list of function arguments.
            self.lbp = self.rbp = 0

    @property
    def source(self) -> str:
        if not self:
            return '?'
        elif len(self) == 1:
            return f'?{self[0].source}'
        else:
            return f'{self[0].source}?{self[1].source}'

    def nud(self) -> 'LookupOperatorToken':
        try:
            self.parser.expected_next('(name)', '(integer)', '(', '*')
        except SyntaxError:
            if self.lbp:
                raise
            return self  # a placeholder/unary lookup token
        else:
            self[:] = self.parser.expression(85),
            return self

    def led(self, left: XPathToken) -> Union['LookupOperatorToken', XPathToken]:
        try:
            self.parser.expected_next('(name)', '(integer)', '(', '*')
        except SyntaxError:
            if isinstance(left.value, str) and is_sequence_type(left.value, self.parser):
                self.lbp = self.rbp = 0
                left.occurrence = '?'
                return left
            raise
        else:
            self[:] = left, self.parser.expression(85)
            return self

    def evaluate(self, context: ta.ContextType = None) -> ta.OneOrMore[ta.ItemType]:
        if not self:
            return self.symbol  # a placeholder token
        return xlist(self.select(context))

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:

        # flatten sequences, don't flatten arrays.
        def flatten(v: ta.ValueType) -> Iterator[ta.ItemType]:
            if isinstance(v, list):
                yield from v
            else:
                yield v

        if not self:
            if isinstance(self.value, list):
                yield from self.value
            else:
                yield self.value
            return

        items: Iterable[ta.ItemType]
        if len(self) == 1:
            # unary lookup operator (used in predicates)
            if context is None:
                raise self.missing_context()
            items = (context.item,)
        else:
            items = self[0].select(context)

        for item in items:
            symbol = self[-1].symbol
            if isinstance(item, XPathMap):
                if symbol == '*':
                    for value in item.values(context):
                        yield from flatten(value)

                elif symbol in ('(name)', '(integer)'):
                    yield from flatten(
                        item(cast(Union[str, int], self[-1].value), context=context)
                    )
                elif symbol == '(':
                    for obj in self[-1].select(context):
                        yield from flatten(item(self.data_value(obj), context=context))

            elif isinstance(item, XPathArray):
                if symbol == '*':
                    for value in item.items(context):
                        yield from flatten(value)
                elif symbol == '(name)':
                    raise self.error('XPTY0004')
                elif symbol == '(integer)':
                    yield from flatten(item(cast(int, self[-1].value), context=context))
                elif symbol == '(':
                    for value in self[-1].select(context):
                        yield from flatten(item(self.data_value(value), context=context))

            elif item == ():
                continue
            else:
                raise self.error('XPTY0004')


XPath31Parser.symbol_table['?'] = LookupOperatorToken


@method('=>', bp=67)
def led__arrow_operator(self: XPathToken, left: XPathToken) -> XPathToken:
    next_token = self.parser.next_token
    if next_token.symbol == '$':
        self[:] = left, self.parser.expression(80)
    elif isinstance(next_token, ProxyToken):
        self.parser.parse_arguments = False
        self[:] = left, next_token.nud()
        self.parser.parse_arguments = True
        self.parser.advance()
    elif isinstance(next_token, XPathFunction):
        self[:] = left, next_token
        if next_token.label == 'kind test':
            raise next_token.wrong_syntax()
        self.parser.advance()  # Skip static evaluation of function arguments
    else:
        next_token.expected('(name)', ':', 'Q{', '(')
        self.parser.parse_arguments = False
        self[:] = left, self.parser.expression(80)
        self.parser.parse_arguments = True

    right = self.parser.expression(67)
    right.expected('(')
    self.append(right)
    return self


@method('=>')
def evaluate__arrow_operator(self: XPathToken, context: ta.ContextType = None) \
        -> ta.ValueType:
    tokens = [self[0]]
    if self[2]:
        tokens.extend(self[2][0].get_argument_tokens())
    func = self[1].get_function(context, arity=len(tokens))
    arguments = [tk.evaluate(context) for tk in tokens]
    return func(*arguments, context=context)



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from abc import abstractmethod
from typing import Union, Any
import codecs

from elementpath.aliases import XPath2ParserType
from elementpath.helpers import LazyPattern, collapse_white_spaces
from .any_types import AnyAtomicType
from .untyped import UntypedAtomic

__all__ = ['AbstractBinary', 'Base64Binary', 'HexBinary']


class AbstractBinary(AnyAtomicType):
    """
    Abstract base class for xs:base64Binary/xs:hexBinary datatypes.
    """
    value: bytes

    __slots__ = ('value', 'ordered')

    @classmethod
    def make(cls, value: Any,
             parser: XPath2ParserType | None = None,
             **kwargs: Any) -> 'AbstractBinary':
        ordered = parser is not None and parser.version >= '3.1'
        return cls(value, ordered)

    @property
    def xpath_versions(self) -> tuple[str, ...] | str:
        return '3.1+' if self.ordered else '2.0', '3.0'

    def __init__(self, value: Union[str, bytes, UntypedAtomic, 'AbstractBinary'],
                 ordered: bool = False) -> None:
        """
        :param value: a string or a binary data or an untyped atomic instance.
        :param ordered: a boolean that enable total ordering for the instance, `False` for default.
        """
        self.ordered = ordered

        match value:
            case x if isinstance(x, self.__class__):
                self.value = value.value
                return
            case AbstractBinary():
                self.value = self.encoder(value.decode())
                return
            case UntypedAtomic():
                value = collapse_white_spaces(value.value)
            case str():
                value = collapse_white_spaces(value)
            case bytes():
                value = collapse_white_spaces(value.decode('utf-8'))
            case _:
                raise self._invalid_type(value)  # noqa

        self.validate(value)
        self.value = value.replace(' ', '').encode('ascii')

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, self.value)

    def __bytes__(self) -> bytes:
        return self.value

    @classmethod
    def validate(cls, value: object) -> None:
        raise NotImplementedError()

    @classmethod
    def encoder(cls, value: bytes) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def decode(self) -> bytes:
        raise NotImplementedError()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, AbstractBinary):
            return self.decode() == other.decode()
        else:
            return NotImplemented

    def __lt__(self, other: object) -> bool:
        if not self.ordered or not isinstance(other, AbstractBinary):
            return NotImplemented

        for oct1, oct2 in zip(self.decode(), other.decode()):
            if oct1 != oct2:
                return oct1 < oct2

        return len(self.decode()) < len(other.decode())

    def __le__(self, other: object) -> bool:
        if not self.ordered or not isinstance(other, AbstractBinary):
            return NotImplemented

        for oct1, oct2 in zip(self.decode(), other.decode()):
            if oct1 != oct2:
                return oct1 < oct2

        return len(self.decode()) <= len(other.decode())

    def __gt__(self, other: object) -> bool:
        if not self.ordered or not isinstance(other, AbstractBinary):
            return NotImplemented

        for oct1, oct2 in zip(self.decode(), other.decode()):
            if oct1 != oct2:
                return oct1 > oct2

        return len(self.decode()) > len(other.decode())

    def __ge__(self, other: object) -> bool:
        if not self.ordered or not isinstance(other, AbstractBinary):
            return NotImplemented

        for oct1, oct2 in zip(self.decode(), other.decode()):
            if oct1 != oct2:
                return oct1 > oct2

        return len(self.decode()) >= len(other.decode())


class Base64Binary(AbstractBinary):
    name = 'base64Binary'
    pattern = LazyPattern(
        r'((?:(?:[A-Za-z0-9+/] ?){4})*(?:(?:[A-Za-z0-9+/] ?){3}[A-Za-z0-9+/]|'
        r'(?:[A-Za-z0-9+/] ?){2}'
        r'[AEIMQUYcgkosw048] ?=|[A-Za-z0-9+/] ?[AQgw] ?= ?=))?'
    )

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, cls):
            return
        elif isinstance(value, bytes):
            value = value.decode()
        elif not isinstance(value, str):
            raise cls._invalid_type(value)

        value = value.replace(' ', '')
        if value:
            match = cls.pattern.match(value)
            if match is None or match.group(0) != value:
                raise cls._invalid_value(value)

    def __str__(self) -> str:
        return self.value.decode('utf-8')

    def __hash__(self) -> int:
        return hash(self.value)

    def __len__(self) -> int:
        length = len(self.value)
        if length == 0:
            return 0
        elif self.value[-2] == ord('='):
            return length // 4 * 3 - 2
        elif self.value[-1] == ord('='):
            return length // 4 * 3 - 1
        return length // 4 * 3

    @classmethod
    def encoder(cls, value: bytes) -> bytes:
        return codecs.encode(value, 'base64').rstrip(b'\n')

    def decode(self) -> bytes:
        return codecs.decode(self.value, 'base64')


class HexBinary(AbstractBinary):
    name = 'hexBinary'
    pattern = LazyPattern(r'^([0-9a-fA-F]{2})*$')

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, cls):
            return
        elif isinstance(value, bytes):
            value = value.decode()
        elif not isinstance(value, str):
            raise cls._invalid_type(value)

        value = value.strip()
        if cls.pattern.match(value) is None:
            raise cls._invalid_value(value)

    @classmethod
    def encoder(cls, value: bytes) -> bytes:
        return codecs.encode(value, 'hex')

    def decode(self) -> bytes:
        return codecs.decode(self.value, 'hex')

    def __str__(self) -> str:
        return self.value.decode('utf-8').upper()

    def __hash__(self) -> int:
        return hash(self.value.upper())

    def __len__(self) -> int:
        return len(self.value) // 2



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import operator
from collections.abc import Callable, Iterable, Sequence
from itertools import chain
from typing import Any, cast, NoReturn, overload, TypeVar, SupportsIndex, Union

from elementpath.aliases import ItemType

__all__ = ['xlist', 'XSequence', 'empty_sequence', 'sequence_concat',
           'sequence_count', 'iterate_sequence']


T = TypeVar('T', bound=ItemType)
_S = TypeVar('_S')


class xlist(list[T]):
    """An extended list type for processing XQuery/XPath sequences."""
    __slots__ = ()

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, list):
            return super().__eq__(other)
        return False if len(self) != 1 else not self[0] != other

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, list):
            return super().__ne__(other)
        return True if len(self) != 1 else not self[0] == other

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, list):
            return super().__lt__(other)
        elif len(self) == 1:
            return bool(self[0] < other)
        return NotImplemented

    def __le__(self, other: Any) -> bool:
        if isinstance(other, list):
            return super().__le__(other)
        elif len(self) == 1:
            return bool(self[0] <= other)
        return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, list):
            return super().__gt__(other)
        elif len(self) == 1:
            return bool(self[0] > other)
        return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, list):
            return super().__ge__(other)
        elif len(self) == 1:
            return bool(self[0] >= other)
        return NotImplemented

    @overload
    def __add__(self, other: list[T]) -> list[T]: ...

    @overload
    def __add__(self, other: list[_S]) -> list[_S | T]: ...

    def __add__(self, other: list[T] | list[_S]) -> list[T] | list[_S | T]:
        if isinstance(other, list):
            return xlist(chain(self, cast(list[T], other)))
        elif len(self) == 1:
            return xlist([self[0] + other])
        return NotImplemented

    def __radd__(self, other: list[T]) -> list[T]:
        if isinstance(other, list):
            return xlist(chain(other, self))
        return NotImplemented

    def __mul__(self, value: SupportsIndex) -> list[T]:
        return NotImplemented

    __rmul__ = __mul__

    @property
    def sequence_type(self) -> str:
        if not self:
            return 'empty-sequence()'
        elif len(self) == 1:
            return 'item()'
        else:
            return 'item()*'


###
# XDM 4.0 constructors and accessors as internal API.
def empty_sequence() -> list[NoReturn]:
    return []


def sequence_concat(s1: list[T], s2: list[T]) -> list[T]:
    match len(s1) + len(s2):
        case 0:
            return []
        case 1:
            return xlist(s1 or s2)
        case _:
            return list(chain(s1, s2))


def sequence_count(seq: list[Any]) -> int:
    return len(seq)


def iterate_sequence(seq: list[T], action: Callable[[T, int], list[T]]) -> list[T]:
    items: list[T] = []
    for position, item in enumerate(seq, start=1):
        items.extend(action(item, position))
    return items if len(items) != 1 else xlist(items)


###
# Sequence class for external usage.
#
SequenceArgType = Union[list[ItemType], tuple[ItemType], 'XSequence']
ActionArgType = Callable[[ItemType, int], list[ItemType]]


class XSequence(Sequence[ItemType]):
    """
    An immutable XPath sequence for external processing of XQuery/XPath sequences.

    Ref: https://qt4cg.org/specifications/xpath-datamodel-40/Overview.html
    """
    __slots__ = '_items',

    @classmethod
    def empty_sequence(cls) -> 'XSequence':
        return cls()

    @classmethod
    def sequence_concat(cls, *seq: SequenceArgType) -> 'XSequence':
        if any(not isinstance(x, (list, tuple, XSequence)) for x in seq):
            raise TypeError("arguments must be sequences")
        return XSequence(chain(*seq))

    @classmethod
    def sequence_count(cls, seq: SequenceArgType) -> int:
        return len(seq)

    @classmethod
    def iterate_sequence(cls, seq: SequenceArgType, action: ActionArgType) -> 'XSequence':
        items: list[ItemType] = []
        for position, item in enumerate(seq, start=1):
            items.extend(action(item, position))
        return XSequence(items)

    def __init__(self, items: Iterable[ItemType] = ()) -> None:
        self._items = tuple(items)
        if any(isinstance(x, (list, XSequence)) for x in self._items):
            raise TypeError(f"{self!r} initialized with a nested sequence")

    def __str__(self) -> str:
        return f'({", ".join(map(repr, self))})'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}([{", ".join(map(repr, self))}])'

    def _operator(self, op: Callable[[Any, Any], bool], other: object) -> bool:
        if isinstance(other, XSequence):
            return op(self._items, other._items)
        elif isinstance(other, tuple):
            return op(self._items, other)
        elif isinstance(other, list):
            return op(self._items, tuple(other))
        elif len(self._items) == 1:
            return op(self._items[0], other)
        else:
            return cast(bool, NotImplemented)

    @overload
    def __getitem__(self, item: int) -> ItemType: ...

    @overload
    def __getitem__(self, item: slice) -> tuple[ItemType, ...]: ...

    def __getitem__(self, item: int | slice) -> ItemType | tuple[ItemType, ...]:
        return self._items[item]

    def __len__(self) -> int:
        return len(self._items)

    def __hash__(self) -> int:
        return hash(self._items)

    def __eq__(self, other: Any) -> bool:
        return self._operator(operator.eq, other)

    def __ne__(self, other: Any) -> bool:
        return not self._operator(operator.eq, other)

    def __lt__(self, other: Any) -> bool:
        return self._operator(operator.lt, other)

    def __le__(self, other: Any) -> bool:
        return self._operator(operator.le, other)

    def __gt__(self, other: Any) -> bool:
        return self._operator(operator.gt, other)

    def __ge__(self, other: Any) -> bool:
        return self._operator(operator.ge, other)

    def __add__(self, other: Any) -> 'XSequence':
        if isinstance(other, (list, tuple, XSequence)):
            return XSequence(chain(self, other))
        elif len(self) == 1:
            return XSequence([self[0] + other])
        return NotImplemented

    def __radd__(self, other: Any) -> 'XSequence':
        if isinstance(other, (list, tuple)):
            return XSequence(chain(other, self))
        return NotImplemented

    @property
    def sequence_type(self) -> str:
        if not self:
            return 'empty-sequence()'
        elif len(self) == 1:
            return 'item()'
        else:
            return 'item()*'



#
# Copyright (c), 2025-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""Experimental use of XDM on pathlib.Path objects."""

from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Optional

from elementpath.aliases import AtomicType, ParentNodeType, ChildNodeType
from elementpath.protocols import XsdElementProtocol
from elementpath.xpath_nodes import XPathNodeTree, AttributeNode, DocumentNode, ElementNode
from elementpath.helpers import match_wildcard
from elementpath.namespaces import XSD_NAMESPACE
from elementpath.datatypes import DateTime


class PathElementNode(ElementNode):
    name: str
    value: Path
    xsd_element: Optional[XsdElementProtocol]

    __slots__ = ('stat',)

    def __init__(self,
                 path: Path,
                 parent: Optional[ParentNodeType] = None,
                 fragment: bool = False) -> None:

        if parent is not None:
            self.parent = parent
            self.tree = parent.tree
        elif fragment:
            self.parent = parent
            path = path.resolve(strict=True)
            self.tree = XPathNodeTree(self, uri=path.as_uri())
        else:
            path = path.resolve(strict=True)

            parents = [p for p in reversed(path.parents)]
            parent = PathDocumentNode(parents[0], uri=path.as_uri())
            self.tree = parent.tree

            for p in parents:
                parent = PathElementNode(p, parent)
            self.parent = parent

        self.name = path.name
        self.value = path
        self.stat = path.stat()
        self.position = self.stat.st_ino
        self.children = []
        self.xsd_type = None
        self._nsmap = None
        self.tree.elements[path] = self

    @property
    def content(self) -> Path:
        return self.value

    elem = content

    @property
    def attributes(self) -> list[AttributeNode]:
        """Stat slots as attributes."""
        if not hasattr(self, '_attributes'):
            self._attributes = []
            for name in dir(self.stat):
                if name[:3] != 'st_':
                    continue
                elif name in ('st_atime_ns', 'st_ctime_ns', 'st_mtime_ns', 'st_birthtime_ns'):
                    dt = datetime.fromtimestamp(getattr(self.stat, name) / 1e9)
                    self._attributes.append(DatetimeAttributeNode(name[3:-3], dt, self))
                elif name[-4:] != 'time':
                    value: int = getattr(self.stat, name)
                    if name == 'st_mode':
                        self._attributes.append(ModeAttributeNode(name[3:], value, self))
                    else:
                        self._attributes.append(IntAttributeNode(name[3:], value, self))
        return self._attributes

    @property
    def name_path(self) -> str:
        return self.name

    @property
    def path(self) -> str:
        if self.parent is None:
            return f'/{self.name_path}'

        if isinstance(self.parent, ElementNode):
            return f"{self.parent.path}/{self.name_path}"
        return f"/{self.name_path}"

    @property
    def string_value(self) -> str:
        if self.value.is_dir():
            return ''
        if self.value.is_file():
            return self.value.read_text()
        else:
            return ''

    @property
    def iter_typed_values(self) -> Iterator[AtomicType]:
        yield from ()

    @property
    def is_typed(self) -> bool:
        return False

    def match_name(self, name: str, default_namespace: Optional[str] = None) -> bool:
        if '*' in name:
            return match_wildcard(self.name, name)
        elif name[0] == '{' or not default_namespace:
            return self.name == name
        else:
            return self.name == f'{{{default_namespace}}}{name}'

    def get_document_node(self, replace: bool = True, as_parent: bool = True) -> 'DocumentNode':
        return PathDocumentNode(Path(self.value.absolute().root))

    def __iter__(self) -> Iterator[ChildNodeType]:
        if not self.children:
            if self.value.is_dir():
                for path in self.value.iterdir():
                    if path in self.tree.elements:
                        self.children.append(self.tree.elements[path])
                    else:
                        self.children.append(PathElementNode(path, self))

        yield from self.children

    def iter_descendants(self, with_self: bool = True) -> Iterator[ChildNodeType]:
        if with_self:
            yield self

        for child in self:
            if isinstance(child, PathElementNode):
                yield from child.iter_descendants()
            else:
                yield child


class PathDocumentNode(DocumentNode):
    value: Path

    __slots__ = ('stat',)

    def __init__(self, document: Path, uri: Optional[str] = None) -> None:

        if not document.is_dir() or len(document.parts) > 1 or not document.is_absolute():
            raise ValueError(f'{document} must be a root directory')

        self.value = document
        self.name = None
        self.parent = None
        self.position = document.stat().st_ino
        self.children = []
        self.tree = XPathNodeTree(self, uri=uri)

    @property
    def document(self) -> Path:
        return self.value

    @property
    def string_value(self) -> str:
        return self.value.read_text()

    def __iter__(self) -> Iterator[ChildNodeType]:
        if not self.children:
            for path in self.value.iterdir():
                if path in self.tree.elements:
                    self.children.append(self.tree.elements[path])
                else:
                    self.children.append(PathElementNode(path, self))

        yield from self.children


class IntAttributeNode(AttributeNode):
    name: str
    value: int
    parent: Optional['PathElementNode']

    __slots__ = ()

    def __init__(self,
                 name: str,
                 value: int,
                 parent: Optional['PathElementNode'] = None) -> None:

        self.name = name
        self.value = value
        self.parent = parent
        self.position = parent.position if parent is not None else 1
        self.xsd_type = None

    @property
    def string_value(self) -> str:
        return str(self.value)

    @property
    def iter_typed_values(self) -> Iterator[AtomicType]:
        yield self.value

    @property
    def type_name(self) -> Optional[str]:
        return f'{{{XSD_NAMESPACE}}}int'


class ModeAttributeNode(IntAttributeNode):

    @property
    def string_value(self) -> str:
        return oct(self.value)


class DatetimeAttributeNode(AttributeNode):
    name: str
    value: datetime
    parent: Optional['PathElementNode']

    __slots__ = ()

    def __init__(self,
                 name: str,
                 value: datetime,
                 parent: Optional['PathElementNode'] = None) -> None:

        self.name = name
        self.value = value
        self.parent = parent
        self.position = parent.position if parent is not None else 1
        self.xsd_type = None

    @property
    def string_value(self) -> str:
        return str(self.value)

    @property
    def iter_typed_values(self) -> Iterator[AtomicType]:
        yield DateTime.fromdatetime(self.value)

    @property
    def type_name(self) -> Optional[str]:
        return f'{{{XSD_NAMESPACE}}}dateTime'
